
import google.protobuf
from packaging.version import Version
if Version(google.protobuf.__version__).major >= 5:
  # -*- coding: utf-8 -*-
  # Generated by the protocol buffer compiler.  DO NOT EDIT!
  # source: webhooks.proto
  # Protobuf Python Version: 5.26.0
  """Generated protocol buffer code."""
  from google.protobuf import descriptor as _descriptor
  from google.protobuf import descriptor_pool as _descriptor_pool
  from google.protobuf import symbol_database as _symbol_database
  from google.protobuf.internal import builder as _builder
  # @@protoc_insertion_point(imports)

  _sym_db = _symbol_database.Default()


  from .scalapb import scalapb_pb2 as scalapb_dot_scalapb__pb2
  from . import databricks_pb2 as databricks__pb2


  DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0ewebhooks.proto\x12\x06mlflow\x1a\x15scalapb/scalapb.proto\x1a\x10\x64\x61tabricks.proto\"h\n\x0cWebhookEvent\x12+\n\x06\x65ntity\x18\x01 \x01(\x0e\x32\x15.mlflow.WebhookEntityB\x04\xf8\x86\x19\x01\x12+\n\x06\x61\x63tion\x18\x02 \x01(\x0e\x32\x15.mlflow.WebhookActionB\x04\xf8\x86\x19\x01\"\xd6\x01\n\x07Webhook\x12\x12\n\nwebhook_id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x0b\n\x03url\x18\x04 \x01(\t\x12$\n\x06\x65vents\x18\x05 \x03(\x0b\x32\x14.mlflow.WebhookEvent\x12%\n\x06status\x18\x06 \x01(\x0e\x32\x15.mlflow.WebhookStatus\x12\x1a\n\x12\x63reation_timestamp\x18\x07 \x01(\x03\x12\x1e\n\x16last_updated_timestamp\x18\x08 \x01(\x03\"k\n\x11WebhookTestResult\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x17\n\x0fresponse_status\x18\x02 \x01(\x05\x12\x15\n\rresponse_body\x18\x03 \x01(\t\x12\x15\n\rerror_message\x18\x04 \x01(\t\"\x89\x02\n\rCreateWebhook\x12\x12\n\x04name\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x12\x11\n\x03url\x18\x03 \x01(\tB\x04\xf8\x86\x19\x01\x12*\n\x06\x65vents\x18\x04 \x03(\x0b\x32\x14.mlflow.WebhookEventB\x04\xf8\x86\x19\x01\x12\x0e\n\x06secret\x18\x05 \x01(\t\x12%\n\x06status\x18\x06 \x01(\x0e\x32\x15.mlflow.WebhookStatus\x1a,\n\x08Response\x12 \n\x07webhook\x18\x01 \x01(\x0b\x32\x0f.mlflow.Webhook:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\xb1\x01\n\x0cListWebhooks\x12\x18\n\x0bmax_results\x18\x01 \x01(\x05:\x03\x31\x30\x30\x12\x12\n\npage_token\x18\x02 \x01(\t\x1a\x46\n\x08Response\x12!\n\x08webhooks\x18\x01 \x03(\x0b\x32\x0f.mlflow.Webhook\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\x81\x01\n\nGetWebhook\x12\x18\n\nwebhook_id\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x1a,\n\x08Response\x12 \n\x07webhook\x18\x01 \x01(\x0b\x32\x0f.mlflow.Webhook:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\x91\x02\n\rUpdateWebhook\x12\x18\n\nwebhook_id\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x0b\n\x03url\x18\x04 \x01(\t\x12$\n\x06\x65vents\x18\x05 \x03(\x0b\x32\x14.mlflow.WebhookEvent\x12\x0e\n\x06secret\x18\x06 \x01(\t\x12%\n\x06status\x18\x07 \x01(\x0e\x32\x15.mlflow.WebhookStatus\x1a,\n\x08Response\x12 \n\x07webhook\x18\x01 \x01(\x0b\x32\x0f.mlflow.Webhook:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"b\n\rDeleteWebhook\x12\x18\n\nwebhook_id\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x1a\n\n\x08Response:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\xb0\x01\n\x0bTestWebhook\x12\x18\n\nwebhook_id\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12#\n\x05\x65vent\x18\x02 \x01(\x0b\x32\x14.mlflow.WebhookEvent\x1a\x35\n\x08Response\x12)\n\x06result\x18\x01 \x01(\x0b\x32\x19.mlflow.WebhookTestResult:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]*)\n\rWebhookStatus\x12\n\n\x06\x41\x43TIVE\x10\x01\x12\x0c\n\x08\x44ISABLED\x10\x02*\xda\x01\n\rWebhookEntity\x12\x16\n\x12\x45NTITY_UNSPECIFIED\x10\x00\x12\x14\n\x10REGISTERED_MODEL\x10\x01\x12\x11\n\rMODEL_VERSION\x10\x02\x12\x15\n\x11MODEL_VERSION_TAG\x10\x03\x12\x17\n\x13MODEL_VERSION_ALIAS\x10\x04\x12\n\n\x06PROMPT\x10\x05\x12\x12\n\x0ePROMPT_VERSION\x10\x06\x12\x0e\n\nPROMPT_TAG\x10\x07\x12\x16\n\x12PROMPT_VERSION_TAG\x10\x08\x12\x10\n\x0cPROMPT_ALIAS\x10\t*W\n\rWebhookAction\x12\x16\n\x12\x41\x43TION_UNSPECIFIED\x10\x00\x12\x0b\n\x07\x43REATED\x10\x01\x12\x0b\n\x07UPDATED\x10\x02\x12\x0b\n\x07\x44\x45LETED\x10\x03\x12\x07\n\x03SET\x10\x04\x32\xb5\x06\n\x0eWebhookService\x12~\n\rcreateWebhook\x12\x15.mlflow.CreateWebhook\x1a\x1e.mlflow.CreateWebhook.Response\"6\xf2\x86\x19\x32\n\x1e\n\x04POST\x12\x10/mlflow/webhooks\x1a\x04\x08\x02\x10\x00\x10\x01*\x0e\x43reate Webhook\x12y\n\x0clistWebhooks\x12\x14.mlflow.ListWebhooks\x1a\x1d.mlflow.ListWebhooks.Response\"4\xf2\x86\x19\x30\n\x1d\n\x03GET\x12\x10/mlflow/webhooks\x1a\x04\x08\x02\x10\x00\x10\x01*\rList Webhooks\x12~\n\ngetWebhook\x12\x12.mlflow.GetWebhook\x1a\x1b.mlflow.GetWebhook.Response\"?\xf2\x86\x19;\n*\n\x03GET\x12\x1d/mlflow/webhooks/{webhook_id}\x1a\x04\x08\x02\x10\x00\x10\x01*\x0bGet Webhook\x12\x8c\x01\n\rupdateWebhook\x12\x15.mlflow.UpdateWebhook\x1a\x1e.mlflow.UpdateWebhook.Response\"D\xf2\x86\x19@\n,\n\x05PATCH\x12\x1d/mlflow/webhooks/{webhook_id}\x1a\x04\x08\x02\x10\x00\x10\x01*\x0eUpdate Webhook\x12\x8d\x01\n\rdeleteWebhook\x12\x15.mlflow.DeleteWebhook\x1a\x1e.mlflow.DeleteWebhook.Response\"E\xf2\x86\x19\x41\n-\n\x06\x44\x45LETE\x12\x1d/mlflow/webhooks/{webhook_id}\x1a\x04\x08\x02\x10\x00\x10\x01*\x0e\x44\x65lete Webhook\x12\x88\x01\n\x0btestWebhook\x12\x13.mlflow.TestWebhook\x1a\x1c.mlflow.TestWebhook.Response\"F\xf2\x86\x19\x42\n0\n\x04POST\x12\"/mlflow/webhooks/{webhook_id}/test\x1a\x04\x08\x02\x10\x00\x10\x01*\x0cTest WebhookB!\n\x14org.mlflow.api.proto\x90\x01\x01\xa0\x01\x01\xe2?\x02\x10\x01')

  _globals = globals()
  _builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
  _builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'webhooks_pb2', _globals)
  if not _descriptor._USE_C_DESCRIPTORS:
    _globals['DESCRIPTOR']._loaded_options = None
    _globals['DESCRIPTOR']._serialized_options = b'\n\024org.mlflow.api.proto\220\001\001\240\001\001\342?\002\020\001'
    _globals['_WEBHOOKEVENT'].fields_by_name['entity']._loaded_options = None
    _globals['_WEBHOOKEVENT'].fields_by_name['entity']._serialized_options = b'\370\206\031\001'
    _globals['_WEBHOOKEVENT'].fields_by_name['action']._loaded_options = None
    _globals['_WEBHOOKEVENT'].fields_by_name['action']._serialized_options = b'\370\206\031\001'
    _globals['_CREATEWEBHOOK'].fields_by_name['name']._loaded_options = None
    _globals['_CREATEWEBHOOK'].fields_by_name['name']._serialized_options = b'\370\206\031\001'
    _globals['_CREATEWEBHOOK'].fields_by_name['url']._loaded_options = None
    _globals['_CREATEWEBHOOK'].fields_by_name['url']._serialized_options = b'\370\206\031\001'
    _globals['_CREATEWEBHOOK'].fields_by_name['events']._loaded_options = None
    _globals['_CREATEWEBHOOK'].fields_by_name['events']._serialized_options = b'\370\206\031\001'
    _globals['_CREATEWEBHOOK']._loaded_options = None
    _globals['_CREATEWEBHOOK']._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
    _globals['_LISTWEBHOOKS']._loaded_options = None
    _globals['_LISTWEBHOOKS']._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
    _globals['_GETWEBHOOK'].fields_by_name['webhook_id']._loaded_options = None
    _globals['_GETWEBHOOK'].fields_by_name['webhook_id']._serialized_options = b'\370\206\031\001'
    _globals['_GETWEBHOOK']._loaded_options = None
    _globals['_GETWEBHOOK']._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
    _globals['_UPDATEWEBHOOK'].fields_by_name['webhook_id']._loaded_options = None
    _globals['_UPDATEWEBHOOK'].fields_by_name['webhook_id']._serialized_options = b'\370\206\031\001'
    _globals['_UPDATEWEBHOOK']._loaded_options = None
    _globals['_UPDATEWEBHOOK']._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
    _globals['_DELETEWEBHOOK'].fields_by_name['webhook_id']._loaded_options = None
    _globals['_DELETEWEBHOOK'].fields_by_name['webhook_id']._serialized_options = b'\370\206\031\001'
    _globals['_DELETEWEBHOOK']._loaded_options = None
    _globals['_DELETEWEBHOOK']._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
    _globals['_TESTWEBHOOK'].fields_by_name['webhook_id']._loaded_options = None
    _globals['_TESTWEBHOOK'].fields_by_name['webhook_id']._serialized_options = b'\370\206\031\001'
    _globals['_TESTWEBHOOK']._loaded_options = None
    _globals['_TESTWEBHOOK']._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
    _globals['_WEBHOOKSERVICE'].methods_by_name['createWebhook']._loaded_options = None
    _globals['_WEBHOOKSERVICE'].methods_by_name['createWebhook']._serialized_options = b'\362\206\0312\n\036\n\004POST\022\020/mlflow/webhooks\032\004\010\002\020\000\020\001*\016Create Webhook'
    _globals['_WEBHOOKSERVICE'].methods_by_name['listWebhooks']._loaded_options = None
    _globals['_WEBHOOKSERVICE'].methods_by_name['listWebhooks']._serialized_options = b'\362\206\0310\n\035\n\003GET\022\020/mlflow/webhooks\032\004\010\002\020\000\020\001*\rList Webhooks'
    _globals['_WEBHOOKSERVICE'].methods_by_name['getWebhook']._loaded_options = None
    _globals['_WEBHOOKSERVICE'].methods_by_name['getWebhook']._serialized_options = b'\362\206\031;\n*\n\003GET\022\035/mlflow/webhooks/{webhook_id}\032\004\010\002\020\000\020\001*\013Get Webhook'
    _globals['_WEBHOOKSERVICE'].methods_by_name['updateWebhook']._loaded_options = None
    _globals['_WEBHOOKSERVICE'].methods_by_name['updateWebhook']._serialized_options = b'\362\206\031@\n,\n\005PATCH\022\035/mlflow/webhooks/{webhook_id}\032\004\010\002\020\000\020\001*\016Update Webhook'
    _globals['_WEBHOOKSERVICE'].methods_by_name['deleteWebhook']._loaded_options = None
    _globals['_WEBHOOKSERVICE'].methods_by_name['deleteWebhook']._serialized_options = b'\362\206\031A\n-\n\006DELETE\022\035/mlflow/webhooks/{webhook_id}\032\004\010\002\020\000\020\001*\016Delete Webhook'
    _globals['_WEBHOOKSERVICE'].methods_by_name['testWebhook']._loaded_options = None
    _globals['_WEBHOOKSERVICE'].methods_by_name['testWebhook']._serialized_options = b'\362\206\031B\n0\n\004POST\022\"/mlflow/webhooks/{webhook_id}/test\032\004\010\002\020\000\020\001*\014Test Webhook'
    _globals['_WEBHOOKSTATUS']._serialized_start=1634
    _globals['_WEBHOOKSTATUS']._serialized_end=1675
    _globals['_WEBHOOKENTITY']._serialized_start=1678
    _globals['_WEBHOOKENTITY']._serialized_end=1896
    _globals['_WEBHOOKACTION']._serialized_start=1898
    _globals['_WEBHOOKACTION']._serialized_end=1985
    _globals['_WEBHOOKEVENT']._serialized_start=67
    _globals['_WEBHOOKEVENT']._serialized_end=171
    _globals['_WEBHOOK']._serialized_start=174
    _globals['_WEBHOOK']._serialized_end=388
    _globals['_WEBHOOKTESTRESULT']._serialized_start=390
    _globals['_WEBHOOKTESTRESULT']._serialized_end=497
    _globals['_CREATEWEBHOOK']._serialized_start=500
    _globals['_CREATEWEBHOOK']._serialized_end=765
    _globals['_CREATEWEBHOOK_RESPONSE']._serialized_start=676
    _globals['_CREATEWEBHOOK_RESPONSE']._serialized_end=720
    _globals['_LISTWEBHOOKS']._serialized_start=768
    _globals['_LISTWEBHOOKS']._serialized_end=945
    _globals['_LISTWEBHOOKS_RESPONSE']._serialized_start=830
    _globals['_LISTWEBHOOKS_RESPONSE']._serialized_end=900
    _globals['_GETWEBHOOK']._serialized_start=948
    _globals['_GETWEBHOOK']._serialized_end=1077
    _globals['_GETWEBHOOK_RESPONSE']._serialized_start=676
    _globals['_GETWEBHOOK_RESPONSE']._serialized_end=720
    _globals['_UPDATEWEBHOOK']._serialized_start=1080
    _globals['_UPDATEWEBHOOK']._serialized_end=1353
    _globals['_UPDATEWEBHOOK_RESPONSE']._serialized_start=676
    _globals['_UPDATEWEBHOOK_RESPONSE']._serialized_end=720
    _globals['_DELETEWEBHOOK']._serialized_start=1355
    _globals['_DELETEWEBHOOK']._serialized_end=1453
    _globals['_DELETEWEBHOOK_RESPONSE']._serialized_start=676
    _globals['_DELETEWEBHOOK_RESPONSE']._serialized_end=686
    _globals['_TESTWEBHOOK']._serialized_start=1456
    _globals['_TESTWEBHOOK']._serialized_end=1632
    _globals['_TESTWEBHOOK_RESPONSE']._serialized_start=1534
    _globals['_TESTWEBHOOK_RESPONSE']._serialized_end=1587
    _globals['_WEBHOOKSERVICE']._serialized_start=1988
    _globals['_WEBHOOKSERVICE']._serialized_end=2809
  _builder.BuildServices(DESCRIPTOR, 'webhooks_pb2', _globals)
  # @@protoc_insertion_point(module_scope)

else:
  # -*- coding: utf-8 -*-
  # Generated by the protocol buffer compiler.  DO NOT EDIT!
  # source: webhooks.proto
  """Generated protocol buffer code."""
  from google.protobuf.internal import enum_type_wrapper
  from google.protobuf import descriptor as _descriptor
  from google.protobuf import descriptor_pool as _descriptor_pool
  from google.protobuf import message as _message
  from google.protobuf import reflection as _reflection
  from google.protobuf import symbol_database as _symbol_database
  from google.protobuf import service as _service
  from google.protobuf import service_reflection
  # @@protoc_insertion_point(imports)

  _sym_db = _symbol_database.Default()


  from .scalapb import scalapb_pb2 as scalapb_dot_scalapb__pb2
  from . import databricks_pb2 as databricks__pb2


  DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0ewebhooks.proto\x12\x06mlflow\x1a\x15scalapb/scalapb.proto\x1a\x10\x64\x61tabricks.proto\"h\n\x0cWebhookEvent\x12+\n\x06\x65ntity\x18\x01 \x01(\x0e\x32\x15.mlflow.WebhookEntityB\x04\xf8\x86\x19\x01\x12+\n\x06\x61\x63tion\x18\x02 \x01(\x0e\x32\x15.mlflow.WebhookActionB\x04\xf8\x86\x19\x01\"\xd6\x01\n\x07Webhook\x12\x12\n\nwebhook_id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x0b\n\x03url\x18\x04 \x01(\t\x12$\n\x06\x65vents\x18\x05 \x03(\x0b\x32\x14.mlflow.WebhookEvent\x12%\n\x06status\x18\x06 \x01(\x0e\x32\x15.mlflow.WebhookStatus\x12\x1a\n\x12\x63reation_timestamp\x18\x07 \x01(\x03\x12\x1e\n\x16last_updated_timestamp\x18\x08 \x01(\x03\"k\n\x11WebhookTestResult\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x17\n\x0fresponse_status\x18\x02 \x01(\x05\x12\x15\n\rresponse_body\x18\x03 \x01(\t\x12\x15\n\rerror_message\x18\x04 \x01(\t\"\x89\x02\n\rCreateWebhook\x12\x12\n\x04name\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x12\x11\n\x03url\x18\x03 \x01(\tB\x04\xf8\x86\x19\x01\x12*\n\x06\x65vents\x18\x04 \x03(\x0b\x32\x14.mlflow.WebhookEventB\x04\xf8\x86\x19\x01\x12\x0e\n\x06secret\x18\x05 \x01(\t\x12%\n\x06status\x18\x06 \x01(\x0e\x32\x15.mlflow.WebhookStatus\x1a,\n\x08Response\x12 \n\x07webhook\x18\x01 \x01(\x0b\x32\x0f.mlflow.Webhook:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\xb1\x01\n\x0cListWebhooks\x12\x18\n\x0bmax_results\x18\x01 \x01(\x05:\x03\x31\x30\x30\x12\x12\n\npage_token\x18\x02 \x01(\t\x1a\x46\n\x08Response\x12!\n\x08webhooks\x18\x01 \x03(\x0b\x32\x0f.mlflow.Webhook\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\x81\x01\n\nGetWebhook\x12\x18\n\nwebhook_id\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x1a,\n\x08Response\x12 \n\x07webhook\x18\x01 \x01(\x0b\x32\x0f.mlflow.Webhook:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\x91\x02\n\rUpdateWebhook\x12\x18\n\nwebhook_id\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x0b\n\x03url\x18\x04 \x01(\t\x12$\n\x06\x65vents\x18\x05 \x03(\x0b\x32\x14.mlflow.WebhookEvent\x12\x0e\n\x06secret\x18\x06 \x01(\t\x12%\n\x06status\x18\x07 \x01(\x0e\x32\x15.mlflow.WebhookStatus\x1a,\n\x08Response\x12 \n\x07webhook\x18\x01 \x01(\x0b\x32\x0f.mlflow.Webhook:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"b\n\rDeleteWebhook\x12\x18\n\nwebhook_id\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x1a\n\n\x08Response:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\xb0\x01\n\x0bTestWebhook\x12\x18\n\nwebhook_id\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12#\n\x05\x65vent\x18\x02 \x01(\x0b\x32\x14.mlflow.WebhookEvent\x1a\x35\n\x08Response\x12)\n\x06result\x18\x01 \x01(\x0b\x32\x19.mlflow.WebhookTestResult:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]*)\n\rWebhookStatus\x12\n\n\x06\x41\x43TIVE\x10\x01\x12\x0c\n\x08\x44ISABLED\x10\x02*\xda\x01\n\rWebhookEntity\x12\x16\n\x12\x45NTITY_UNSPECIFIED\x10\x00\x12\x14\n\x10REGISTERED_MODEL\x10\x01\x12\x11\n\rMODEL_VERSION\x10\x02\x12\x15\n\x11MODEL_VERSION_TAG\x10\x03\x12\x17\n\x13MODEL_VERSION_ALIAS\x10\x04\x12\n\n\x06PROMPT\x10\x05\x12\x12\n\x0ePROMPT_VERSION\x10\x06\x12\x0e\n\nPROMPT_TAG\x10\x07\x12\x16\n\x12PROMPT_VERSION_TAG\x10\x08\x12\x10\n\x0cPROMPT_ALIAS\x10\t*W\n\rWebhookAction\x12\x16\n\x12\x41\x43TION_UNSPECIFIED\x10\x00\x12\x0b\n\x07\x43REATED\x10\x01\x12\x0b\n\x07UPDATED\x10\x02\x12\x0b\n\x07\x44\x45LETED\x10\x03\x12\x07\n\x03SET\x10\x04\x32\xb5\x06\n\x0eWebhookService\x12~\n\rcreateWebhook\x12\x15.mlflow.CreateWebhook\x1a\x1e.mlflow.CreateWebhook.Response\"6\xf2\x86\x19\x32\n\x1e\n\x04POST\x12\x10/mlflow/webhooks\x1a\x04\x08\x02\x10\x00\x10\x01*\x0e\x43reate Webhook\x12y\n\x0clistWebhooks\x12\x14.mlflow.ListWebhooks\x1a\x1d.mlflow.ListWebhooks.Response\"4\xf2\x86\x19\x30\n\x1d\n\x03GET\x12\x10/mlflow/webhooks\x1a\x04\x08\x02\x10\x00\x10\x01*\rList Webhooks\x12~\n\ngetWebhook\x12\x12.mlflow.GetWebhook\x1a\x1b.mlflow.GetWebhook.Response\"?\xf2\x86\x19;\n*\n\x03GET\x12\x1d/mlflow/webhooks/{webhook_id}\x1a\x04\x08\x02\x10\x00\x10\x01*\x0bGet Webhook\x12\x8c\x01\n\rupdateWebhook\x12\x15.mlflow.UpdateWebhook\x1a\x1e.mlflow.UpdateWebhook.Response\"D\xf2\x86\x19@\n,\n\x05PATCH\x12\x1d/mlflow/webhooks/{webhook_id}\x1a\x04\x08\x02\x10\x00\x10\x01*\x0eUpdate Webhook\x12\x8d\x01\n\rdeleteWebhook\x12\x15.mlflow.DeleteWebhook\x1a\x1e.mlflow.DeleteWebhook.Response\"E\xf2\x86\x19\x41\n-\n\x06\x44\x45LETE\x12\x1d/mlflow/webhooks/{webhook_id}\x1a\x04\x08\x02\x10\x00\x10\x01*\x0e\x44\x65lete Webhook\x12\x88\x01\n\x0btestWebhook\x12\x13.mlflow.TestWebhook\x1a\x1c.mlflow.TestWebhook.Response\"F\xf2\x86\x19\x42\n0\n\x04POST\x12\"/mlflow/webhooks/{webhook_id}/test\x1a\x04\x08\x02\x10\x00\x10\x01*\x0cTest WebhookB!\n\x14org.mlflow.api.proto\x90\x01\x01\xa0\x01\x01\xe2?\x02\x10\x01')

  _WEBHOOKSTATUS = DESCRIPTOR.enum_types_by_name['WebhookStatus']
  WebhookStatus = enum_type_wrapper.EnumTypeWrapper(_WEBHOOKSTATUS)
  _WEBHOOKENTITY = DESCRIPTOR.enum_types_by_name['WebhookEntity']
  WebhookEntity = enum_type_wrapper.EnumTypeWrapper(_WEBHOOKENTITY)
  _WEBHOOKACTION = DESCRIPTOR.enum_types_by_name['WebhookAction']
  WebhookAction = enum_type_wrapper.EnumTypeWrapper(_WEBHOOKACTION)
  ACTIVE = 1
  DISABLED = 2
  ENTITY_UNSPECIFIED = 0
  REGISTERED_MODEL = 1
  MODEL_VERSION = 2
  MODEL_VERSION_TAG = 3
  MODEL_VERSION_ALIAS = 4
  PROMPT = 5
  PROMPT_VERSION = 6
  PROMPT_TAG = 7
  PROMPT_VERSION_TAG = 8
  PROMPT_ALIAS = 9
  ACTION_UNSPECIFIED = 0
  CREATED = 1
  UPDATED = 2
  DELETED = 3
  SET = 4


  _WEBHOOKEVENT = DESCRIPTOR.message_types_by_name['WebhookEvent']
  _WEBHOOK = DESCRIPTOR.message_types_by_name['Webhook']
  _WEBHOOKTESTRESULT = DESCRIPTOR.message_types_by_name['WebhookTestResult']
  _CREATEWEBHOOK = DESCRIPTOR.message_types_by_name['CreateWebhook']
  _CREATEWEBHOOK_RESPONSE = _CREATEWEBHOOK.nested_types_by_name['Response']
  _LISTWEBHOOKS = DESCRIPTOR.message_types_by_name['ListWebhooks']
  _LISTWEBHOOKS_RESPONSE = _LISTWEBHOOKS.nested_types_by_name['Response']
  _GETWEBHOOK = DESCRIPTOR.message_types_by_name['GetWebhook']
  _GETWEBHOOK_RESPONSE = _GETWEBHOOK.nested_types_by_name['Response']
  _UPDATEWEBHOOK = DESCRIPTOR.message_types_by_name['UpdateWebhook']
  _UPDATEWEBHOOK_RESPONSE = _UPDATEWEBHOOK.nested_types_by_name['Response']
  _DELETEWEBHOOK = DESCRIPTOR.message_types_by_name['DeleteWebhook']
  _DELETEWEBHOOK_RESPONSE = _DELETEWEBHOOK.nested_types_by_name['Response']
  _TESTWEBHOOK = DESCRIPTOR.message_types_by_name['TestWebhook']
  _TESTWEBHOOK_RESPONSE = _TESTWEBHOOK.nested_types_by_name['Response']
  WebhookEvent = _reflection.GeneratedProtocolMessageType('WebhookEvent', (_message.Message,), {
    'DESCRIPTOR' : _WEBHOOKEVENT,
    '__module__' : 'webhooks_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.WebhookEvent)
    })
  _sym_db.RegisterMessage(WebhookEvent)

  Webhook = _reflection.GeneratedProtocolMessageType('Webhook', (_message.Message,), {
    'DESCRIPTOR' : _WEBHOOK,
    '__module__' : 'webhooks_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.Webhook)
    })
  _sym_db.RegisterMessage(Webhook)

  WebhookTestResult = _reflection.GeneratedProtocolMessageType('WebhookTestResult', (_message.Message,), {
    'DESCRIPTOR' : _WEBHOOKTESTRESULT,
    '__module__' : 'webhooks_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.WebhookTestResult)
    })
  _sym_db.RegisterMessage(WebhookTestResult)

  CreateWebhook = _reflection.GeneratedProtocolMessageType('CreateWebhook', (_message.Message,), {

    'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
      'DESCRIPTOR' : _CREATEWEBHOOK_RESPONSE,
      '__module__' : 'webhooks_pb2'
      # @@protoc_insertion_point(class_scope:mlflow.CreateWebhook.Response)
      })
    ,
    'DESCRIPTOR' : _CREATEWEBHOOK,
    '__module__' : 'webhooks_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.CreateWebhook)
    })
  _sym_db.RegisterMessage(CreateWebhook)
  _sym_db.RegisterMessage(CreateWebhook.Response)

  ListWebhooks = _reflection.GeneratedProtocolMessageType('ListWebhooks', (_message.Message,), {

    'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
      'DESCRIPTOR' : _LISTWEBHOOKS_RESPONSE,
      '__module__' : 'webhooks_pb2'
      # @@protoc_insertion_point(class_scope:mlflow.ListWebhooks.Response)
      })
    ,
    'DESCRIPTOR' : _LISTWEBHOOKS,
    '__module__' : 'webhooks_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.ListWebhooks)
    })
  _sym_db.RegisterMessage(ListWebhooks)
  _sym_db.RegisterMessage(ListWebhooks.Response)

  GetWebhook = _reflection.GeneratedProtocolMessageType('GetWebhook', (_message.Message,), {

    'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
      'DESCRIPTOR' : _GETWEBHOOK_RESPONSE,
      '__module__' : 'webhooks_pb2'
      # @@protoc_insertion_point(class_scope:mlflow.GetWebhook.Response)
      })
    ,
    'DESCRIPTOR' : _GETWEBHOOK,
    '__module__' : 'webhooks_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.GetWebhook)
    })
  _sym_db.RegisterMessage(GetWebhook)
  _sym_db.RegisterMessage(GetWebhook.Response)

  UpdateWebhook = _reflection.GeneratedProtocolMessageType('UpdateWebhook', (_message.Message,), {

    'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
      'DESCRIPTOR' : _UPDATEWEBHOOK_RESPONSE,
      '__module__' : 'webhooks_pb2'
      # @@protoc_insertion_point(class_scope:mlflow.UpdateWebhook.Response)
      })
    ,
    'DESCRIPTOR' : _UPDATEWEBHOOK,
    '__module__' : 'webhooks_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.UpdateWebhook)
    })
  _sym_db.RegisterMessage(UpdateWebhook)
  _sym_db.RegisterMessage(UpdateWebhook.Response)

  DeleteWebhook = _reflection.GeneratedProtocolMessageType('DeleteWebhook', (_message.Message,), {

    'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
      'DESCRIPTOR' : _DELETEWEBHOOK_RESPONSE,
      '__module__' : 'webhooks_pb2'
      # @@protoc_insertion_point(class_scope:mlflow.DeleteWebhook.Response)
      })
    ,
    'DESCRIPTOR' : _DELETEWEBHOOK,
    '__module__' : 'webhooks_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.DeleteWebhook)
    })
  _sym_db.RegisterMessage(DeleteWebhook)
  _sym_db.RegisterMessage(DeleteWebhook.Response)

  TestWebhook = _reflection.GeneratedProtocolMessageType('TestWebhook', (_message.Message,), {

    'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
      'DESCRIPTOR' : _TESTWEBHOOK_RESPONSE,
      '__module__' : 'webhooks_pb2'
      # @@protoc_insertion_point(class_scope:mlflow.TestWebhook.Response)
      })
    ,
    'DESCRIPTOR' : _TESTWEBHOOK,
    '__module__' : 'webhooks_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.TestWebhook)
    })
  _sym_db.RegisterMessage(TestWebhook)
  _sym_db.RegisterMessage(TestWebhook.Response)

  _WEBHOOKSERVICE = DESCRIPTOR.services_by_name['WebhookService']
  if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b'\n\024org.mlflow.api.proto\220\001\001\240\001\001\342?\002\020\001'
    _WEBHOOKEVENT.fields_by_name['entity']._options = None
    _WEBHOOKEVENT.fields_by_name['entity']._serialized_options = b'\370\206\031\001'
    _WEBHOOKEVENT.fields_by_name['action']._options = None
    _WEBHOOKEVENT.fields_by_name['action']._serialized_options = b'\370\206\031\001'
    _CREATEWEBHOOK.fields_by_name['name']._options = None
    _CREATEWEBHOOK.fields_by_name['name']._serialized_options = b'\370\206\031\001'
    _CREATEWEBHOOK.fields_by_name['url']._options = None
    _CREATEWEBHOOK.fields_by_name['url']._serialized_options = b'\370\206\031\001'
    _CREATEWEBHOOK.fields_by_name['events']._options = None
    _CREATEWEBHOOK.fields_by_name['events']._serialized_options = b'\370\206\031\001'
    _CREATEWEBHOOK._options = None
    _CREATEWEBHOOK._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
    _LISTWEBHOOKS._options = None
    _LISTWEBHOOKS._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
    _GETWEBHOOK.fields_by_name['webhook_id']._options = None
    _GETWEBHOOK.fields_by_name['webhook_id']._serialized_options = b'\370\206\031\001'
    _GETWEBHOOK._options = None
    _GETWEBHOOK._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
    _UPDATEWEBHOOK.fields_by_name['webhook_id']._options = None
    _UPDATEWEBHOOK.fields_by_name['webhook_id']._serialized_options = b'\370\206\031\001'
    _UPDATEWEBHOOK._options = None
    _UPDATEWEBHOOK._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
    _DELETEWEBHOOK.fields_by_name['webhook_id']._options = None
    _DELETEWEBHOOK.fields_by_name['webhook_id']._serialized_options = b'\370\206\031\001'
    _DELETEWEBHOOK._options = None
    _DELETEWEBHOOK._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
    _TESTWEBHOOK.fields_by_name['webhook_id']._options = None
    _TESTWEBHOOK.fields_by_name['webhook_id']._serialized_options = b'\370\206\031\001'
    _TESTWEBHOOK._options = None
    _TESTWEBHOOK._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
    _WEBHOOKSERVICE.methods_by_name['createWebhook']._options = None
    _WEBHOOKSERVICE.methods_by_name['createWebhook']._serialized_options = b'\362\206\0312\n\036\n\004POST\022\020/mlflow/webhooks\032\004\010\002\020\000\020\001*\016Create Webhook'
    _WEBHOOKSERVICE.methods_by_name['listWebhooks']._options = None
    _WEBHOOKSERVICE.methods_by_name['listWebhooks']._serialized_options = b'\362\206\0310\n\035\n\003GET\022\020/mlflow/webhooks\032\004\010\002\020\000\020\001*\rList Webhooks'
    _WEBHOOKSERVICE.methods_by_name['getWebhook']._options = None
    _WEBHOOKSERVICE.methods_by_name['getWebhook']._serialized_options = b'\362\206\031;\n*\n\003GET\022\035/mlflow/webhooks/{webhook_id}\032\004\010\002\020\000\020\001*\013Get Webhook'
    _WEBHOOKSERVICE.methods_by_name['updateWebhook']._options = None
    _WEBHOOKSERVICE.methods_by_name['updateWebhook']._serialized_options = b'\362\206\031@\n,\n\005PATCH\022\035/mlflow/webhooks/{webhook_id}\032\004\010\002\020\000\020\001*\016Update Webhook'
    _WEBHOOKSERVICE.methods_by_name['deleteWebhook']._options = None
    _WEBHOOKSERVICE.methods_by_name['deleteWebhook']._serialized_options = b'\362\206\031A\n-\n\006DELETE\022\035/mlflow/webhooks/{webhook_id}\032\004\010\002\020\000\020\001*\016Delete Webhook'
    _WEBHOOKSERVICE.methods_by_name['testWebhook']._options = None
    _WEBHOOKSERVICE.methods_by_name['testWebhook']._serialized_options = b'\362\206\031B\n0\n\004POST\022\"/mlflow/webhooks/{webhook_id}/test\032\004\010\002\020\000\020\001*\014Test Webhook'
    _WEBHOOKSTATUS._serialized_start=1634
    _WEBHOOKSTATUS._serialized_end=1675
    _WEBHOOKENTITY._serialized_start=1678
    _WEBHOOKENTITY._serialized_end=1896
    _WEBHOOKACTION._serialized_start=1898
    _WEBHOOKACTION._serialized_end=1985
    _WEBHOOKEVENT._serialized_start=67
    _WEBHOOKEVENT._serialized_end=171
    _WEBHOOK._serialized_start=174
    _WEBHOOK._serialized_end=388
    _WEBHOOKTESTRESULT._serialized_start=390
    _WEBHOOKTESTRESULT._serialized_end=497
    _CREATEWEBHOOK._serialized_start=500
    _CREATEWEBHOOK._serialized_end=765
    _CREATEWEBHOOK_RESPONSE._serialized_start=676
    _CREATEWEBHOOK_RESPONSE._serialized_end=720
    _LISTWEBHOOKS._serialized_start=768
    _LISTWEBHOOKS._serialized_end=945
    _LISTWEBHOOKS_RESPONSE._serialized_start=830
    _LISTWEBHOOKS_RESPONSE._serialized_end=900
    _GETWEBHOOK._serialized_start=948
    _GETWEBHOOK._serialized_end=1077
    _GETWEBHOOK_RESPONSE._serialized_start=676
    _GETWEBHOOK_RESPONSE._serialized_end=720
    _UPDATEWEBHOOK._serialized_start=1080
    _UPDATEWEBHOOK._serialized_end=1353
    _UPDATEWEBHOOK_RESPONSE._serialized_start=676
    _UPDATEWEBHOOK_RESPONSE._serialized_end=720
    _DELETEWEBHOOK._serialized_start=1355
    _DELETEWEBHOOK._serialized_end=1453
    _DELETEWEBHOOK_RESPONSE._serialized_start=676
    _DELETEWEBHOOK_RESPONSE._serialized_end=686
    _TESTWEBHOOK._serialized_start=1456
    _TESTWEBHOOK._serialized_end=1632
    _TESTWEBHOOK_RESPONSE._serialized_start=1534
    _TESTWEBHOOK_RESPONSE._serialized_end=1587
    _WEBHOOKSERVICE._serialized_start=1988
    _WEBHOOKSERVICE._serialized_end=2809
  WebhookService = service_reflection.GeneratedServiceType('WebhookService', (_service.Service,), dict(
    DESCRIPTOR = _WEBHOOKSERVICE,
    __module__ = 'webhooks_pb2'
    ))

  WebhookService_Stub = service_reflection.GeneratedServiceStubType('WebhookService_Stub', (WebhookService,), dict(
    DESCRIPTOR = _WEBHOOKSERVICE,
    __module__ = 'webhooks_pb2'
    ))


  # @@protoc_insertion_point(module_scope)




import json
import os
import posixpath
import urllib.parse
from datetime import datetime, timezone
from functools import lru_cache
from mimetypes import guess_type

from mlflow.entities import FileInfo
from mlflow.entities.multipart_upload import (
    CreateMultipartUploadResponse,
    MultipartUploadCredential,
)
from mlflow.environment_variables import (
    MLFLOW_BOTO_CLIENT_ADDRESSING_STYLE,
    MLFLOW_S3_ENDPOINT_URL,
    MLFLOW_S3_IGNORE_TLS,
    MLFLOW_S3_UPLOAD_EXTRA_ARGS,
)
from mlflow.exceptions import MlflowException
from mlflow.protos.databricks_pb2 import (
    INTERNAL_ERROR,
    PERMISSION_DENIED,
    RESOURCE_DOES_NOT_EXIST,
    UNAUTHENTICATED,
)
from mlflow.store.artifact.artifact_repo import (
    ArtifactRepository,
    MultipartUploadMixin,
)
from mlflow.utils.file_utils import relative_path_to_artifact_path

_MAX_CACHE_SECONDS = 300

BOTO_TO_MLFLOW_ERROR = {
    "AccessDenied": PERMISSION_DENIED,
    "NoSuchBucket": RESOURCE_DOES_NOT_EXIST,
    "NoSuchKey": RESOURCE_DOES_NOT_EXIST,
    "InvalidAccessKeyId": UNAUTHENTICATED,
    "SignatureDoesNotMatch": UNAUTHENTICATED,
}


def _get_utcnow_timestamp():
    return datetime.now(timezone.utc).timestamp()


@lru_cache(maxsize=64)
def _cached_get_s3_client(
    signature_version,
    addressing_style,
    s3_endpoint_url,
    verify,
    timestamp,
    access_key_id=None,
    secret_access_key=None,
    session_token=None,
    region_name=None,
):
    """Returns a boto3 client, caching to avoid extra boto3 verify calls.

    This method is outside of the S3ArtifactRepository as it is
    agnostic and could be used by other instances.

    `maxsize` set to avoid excessive memory consumption in the case
    a user has dynamic endpoints (intentionally or as a bug).

    Some of the boto3 endpoint urls, in very edge cases, might expire
    after twelve hours as that is the current expiration time. To ensure
    we throw an error on verification instead of using an expired endpoint
    we utilise the `timestamp` parameter to invalidate cache.
    """
    import boto3
    from botocore.client import Config

    # Making it possible to access public S3 buckets
    # Workaround for https://github.com/boto/botocore/issues/2442
    if signature_version.lower() == "unsigned":
        from botocore import UNSIGNED

        signature_version = UNSIGNED

    return boto3.client(
        "s3",
        config=Config(
            signature_version=signature_version, s3={"addressing_style": addressing_style}
        ),
        endpoint_url=s3_endpoint_url,
        verify=verify,
        aws_access_key_id=access_key_id,
        aws_secret_access_key=secret_access_key,
        aws_session_token=session_token,
        region_name=region_name,
    )


def _get_s3_client(
    addressing_style=None,
    access_key_id=None,
    secret_access_key=None,
    session_token=None,
    region_name=None,
    s3_endpoint_url=None,
):
    if not s3_endpoint_url:
        s3_endpoint_url = MLFLOW_S3_ENDPOINT_URL.get()
    do_verify = not MLFLOW_S3_IGNORE_TLS.get()

    # The valid verify argument value is None/False/path to cert bundle file, See
    # https://github.com/boto/boto3/blob/73865126cad3938ca80a2f567a1c79cb248169a7/
    # boto3/session.py#L212
    verify = None if do_verify else False

    # NOTE: If you need to specify this env variable, please file an issue at
    # https://github.com/mlflow/mlflow/issues so we know your use-case!
    signature_version = os.environ.get("MLFLOW_EXPERIMENTAL_S3_SIGNATURE_VERSION", "s3v4")

    # Invalidate cache every `_MAX_CACHE_SECONDS`
    timestamp = int(_get_utcnow_timestamp() / _MAX_CACHE_SECONDS)

    if not addressing_style:
        addressing_style = MLFLOW_BOTO_CLIENT_ADDRESSING_STYLE.get()

    return _cached_get_s3_client(
        signature_version,
        addressing_style,
        s3_endpoint_url,
        verify,
        timestamp,
        access_key_id=access_key_id,
        secret_access_key=secret_access_key,
        session_token=session_token,
        region_name=region_name,
    )


class S3ArtifactRepository(ArtifactRepository, MultipartUploadMixin):
    """
    Stores artifacts on Amazon S3.

    This repository provides MLflow artifact storage using Amazon S3 as the backend.
    It supports both single-file uploads and multipart uploads for large files,
    with automatic content type detection and configurable upload parameters.

    The repository uses boto3 for S3 operations and supports various authentication
    methods including AWS credentials, IAM roles, and environment variables.

    Environment Variables:
        AWS_ACCESS_KEY_ID: AWS access key ID for authentication
        AWS_SECRET_ACCESS_KEY: AWS secret access key for authentication
        AWS_SESSION_TOKEN: AWS session token for temporary credentials
        AWS_DEFAULT_REGION: Default AWS region for S3 operations
        MLFLOW_S3_ENDPOINT_URL: Custom S3 endpoint URL (for S3-compatible storage)
        MLFLOW_S3_IGNORE_TLS: Set to 'true' to disable TLS verification
        MLFLOW_S3_UPLOAD_EXTRA_ARGS: JSON string of extra arguments for S3 uploads
        MLFLOW_BOTO_CLIENT_ADDRESSING_STYLE: S3 addressing style ('path' or 'virtual')

    Note:
        This class inherits from both ArtifactRepository and MultipartUploadMixin,
        providing full artifact management capabilities including efficient large file uploads.
    """

    def __init__(
        self,
        artifact_uri: str,
        access_key_id=None,
        secret_access_key=None,
        session_token=None,
        tracking_uri: str | None = None,
        registry_uri: str | None = None,
    ) -> None:
        """
        Initialize an S3 artifact repository.

        Args:
            artifact_uri: S3 URI in the format 's3://bucket-name/path/to/artifacts'.
                The URI must be a valid S3 URI with a bucket that exists and is accessible.
            access_key_id: Optional AWS access key ID. If None, uses default AWS credential
                resolution (environment variables, IAM roles, etc.).
            secret_access_key: Optional AWS secret access key. Must be provided if
                access_key_id is provided.
            session_token: Optional AWS session token for temporary credentials.
                Used with STS tokens or IAM roles.
            tracking_uri: Optional URI for the MLflow tracking server.
                If None, uses the current tracking URI context.
        """
        super().__init__(artifact_uri, tracking_uri, registry_uri)
        self._access_key_id = access_key_id
        self._secret_access_key = secret_access_key
        self._session_token = session_token

    def _get_s3_client(self):
        return _get_s3_client(
            access_key_id=self._access_key_id,
            secret_access_key=self._secret_access_key,
            session_token=self._session_token,
        )

    def parse_s3_compliant_uri(self, uri):
        """
        Parse an S3 URI into bucket and path components.

        Args:
            uri: S3 URI in the format 's3://bucket-name/path/to/object'

        Returns:
            A tuple containing (bucket_name, object_path) where:
            - bucket_name: The S3 bucket name
            - object_path: The path within the bucket (without leading slash)
        """
        parsed = urllib.parse.urlparse(uri)
        if parsed.scheme != "s3":
            raise Exception(f"Not an S3 URI: {uri}")
        path = parsed.path
        if path.startswith("/"):
            path = path[1:]
        return parsed.netloc, path

    @staticmethod
    def get_s3_file_upload_extra_args():
        """
        Get additional S3 upload arguments from environment variables.

        Returns:
            Dictionary of extra arguments for S3 uploads, or None if not configured.
            These arguments are passed to boto3's upload_file method.

        Environment Variables:
            MLFLOW_S3_UPLOAD_EXTRA_ARGS: JSON string containing extra arguments
                for S3 uploads (e.g., '{"ServerSideEncryption": "AES256"}')
        """
        s3_file_upload_extra_args = MLFLOW_S3_UPLOAD_EXTRA_ARGS.get()
        if s3_file_upload_extra_args:
            return json.loads(s3_file_upload_extra_args)
        else:
            return None

    def _upload_file(self, s3_client, local_file, bucket, key):
        extra_args = {}
        guessed_type, guessed_encoding = guess_type(local_file)
        if guessed_type is not None:
            extra_args["ContentType"] = guessed_type
        if guessed_encoding is not None:
            extra_args["ContentEncoding"] = guessed_encoding
        environ_extra_args = self.get_s3_file_upload_extra_args()
        if environ_extra_args is not None:
            extra_args.update(environ_extra_args)
        s3_client.upload_file(Filename=local_file, Bucket=bucket, Key=key, ExtraArgs=extra_args)

    def log_artifact(self, local_file, artifact_path=None):
        """
        Log a local file as an artifact to S3.

        This method uploads a single file to S3 with automatic content type detection
        and optional extra upload arguments from environment variables.

        Args:
            local_file: Absolute path to the local file to upload. The file must
                exist and be readable.
            artifact_path: Optional relative path within the S3 bucket where the
                artifact should be stored. If None, the file is stored in the root
                of the configured S3 path. Use forward slashes (/) for path separators.
        """
        (bucket, dest_path) = self.parse_s3_compliant_uri(self.artifact_uri)
        if artifact_path:
            dest_path = posixpath.join(dest_path, artifact_path)
        dest_path = posixpath.join(dest_path, os.path.basename(local_file))
        self._upload_file(
            s3_client=self._get_s3_client(), local_file=local_file, bucket=bucket, key=dest_path
        )

    def log_artifacts(self, local_dir, artifact_path=None):
        """
        Log all files in a local directory as artifacts to S3.

        This method recursively uploads all files in the specified directory,
        preserving the directory structure in S3. Each file is uploaded with
        automatic content type detection.

        Args:
            local_dir: Absolute path to the local directory containing files to upload.
                The directory must exist and be readable.
            artifact_path: Optional relative path within the S3 bucket where the
                artifacts should be stored. If None, files are stored in the root
                of the configured S3 path. Use forward slashes (/) for path separators.
        """
        (bucket, dest_path) = self.parse_s3_compliant_uri(self.artifact_uri)
        if artifact_path:
            dest_path = posixpath.join(dest_path, artifact_path)
        s3_client = self._get_s3_client()
        local_dir = os.path.abspath(local_dir)
        for root, _, filenames in os.walk(local_dir):
            upload_path = dest_path
            if root != local_dir:
                rel_path = os.path.relpath(root, local_dir)
                rel_path = relative_path_to_artifact_path(rel_path)
                upload_path = posixpath.join(dest_path, rel_path)

            for f in filenames:
                self._upload_file(
                    s3_client=s3_client,
                    local_file=os.path.join(root, f),
                    bucket=bucket,
                    key=posixpath.join(upload_path, f),
                )

    def _iterate_s3_paginated_results(self, bucket, prefix):
        """
        Iterate over paginated S3 list_objects_v2 results with error handling.

        This helper method isolates the S3 client operations that can raise ClientError
        and provides appropriate error handling and mapping to MLflow exceptions.
        The ClientError can occur during both paginator setup and iteration, because
        the botocore library makes lazy calls.

        Args:
            bucket: S3 bucket name
            prefix: S3 prefix to list objects under

        Yields:
            Individual result pages from S3 list_objects_v2 operation

        Raises:
            MlflowException: If S3 client operations fail
        """
        from botocore.exceptions import ClientError

        try:
            s3_client = self._get_s3_client()
            paginator = s3_client.get_paginator("list_objects_v2")
            results = paginator.paginate(Bucket=bucket, Prefix=prefix, Delimiter="/")
            for result in results:
                yield result
        except ClientError as error:
            error_code = error.response["Error"]["Code"]
            mlflow_error_code = BOTO_TO_MLFLOW_ERROR.get(error_code, INTERNAL_ERROR)
            error_message = error.response["Error"]["Message"]
            raise MlflowException(
                f"Failed to list artifacts in {self.artifact_uri}: {error_message}",
                error_code=mlflow_error_code,
            )

    def list_artifacts(self, path=None):
        """
        List all artifacts directly under the specified S3 path.

        This method uses S3's list_objects_v2 API with pagination to efficiently
        list artifacts. It treats S3 prefixes as directories and returns both
        files and directories as FileInfo objects.

        Args:
            path: Optional relative path within the S3 bucket to list. If None,
                lists artifacts in the root of the configured S3 path. If the path
                refers to a single file, returns an empty list per MLflow convention.

        Returns:
            A list of FileInfo objects representing artifacts directly under the
            specified path. Each FileInfo contains:
            - path: Relative path of the artifact from the repository root
            - is_dir: True if the artifact represents a directory (S3 prefix)
            - file_size: Size in bytes for files, None for directories
        """
        (bucket, artifact_path) = self.parse_s3_compliant_uri(self.artifact_uri)
        dest_path = artifact_path
        if path:
            dest_path = posixpath.join(dest_path, path)
        dest_path = dest_path.rstrip("/") if dest_path else ""
        infos = []
        prefix = dest_path + "/" if dest_path else ""
        for result in self._iterate_s3_paginated_results(bucket, prefix):
            # Subdirectories will be listed as "common prefixes"
            # due to the way we made the request
            for obj in result.get("CommonPrefixes", []):
                subdir_path = obj.get("Prefix")
                self._verify_listed_object_contains_artifact_path_prefix(
                    listed_object_path=subdir_path, artifact_path=artifact_path
                )
                subdir_rel_path = posixpath.relpath(path=subdir_path, start=artifact_path)
                if subdir_rel_path.endswith("/"):
                    subdir_rel_path = subdir_rel_path[:-1]
                infos.append(FileInfo(subdir_rel_path, True, None))
            # Objects listed directly will be files
            for obj in result.get("Contents", []):
                file_path = obj.get("Key")
                self._verify_listed_object_contains_artifact_path_prefix(
                    listed_object_path=file_path, artifact_path=artifact_path
                )
                file_rel_path = posixpath.relpath(path=file_path, start=artifact_path)
                file_size = int(obj.get("Size"))
                infos.append(FileInfo(file_rel_path, False, file_size))
        return sorted(infos, key=lambda f: f.path)

    @staticmethod
    def _verify_listed_object_contains_artifact_path_prefix(listed_object_path, artifact_path):
        if not listed_object_path.startswith(artifact_path):
            raise MlflowException(
                "The path of the listed S3 object does not begin with the specified"
                f" artifact path. Artifact path: {artifact_path}. Object path:"
                f" {listed_object_path}."
            )

    def _download_file(self, remote_file_path, local_path):
        """
        Download a file from S3 to the local filesystem.

        This method downloads a single file from S3 to the specified local path.
        It's used internally by the download_artifacts method.

        Args:
            remote_file_path: Relative path of the file within the S3 bucket,
                relative to the repository's root path.
            local_path: Absolute path where the file should be saved locally.
                The parent directory must exist.
        """
        (bucket, s3_root_path) = self.parse_s3_compliant_uri(self.artifact_uri)
        s3_full_path = posixpath.join(s3_root_path, remote_file_path)
        s3_client = self._get_s3_client()
        s3_client.download_file(bucket, s3_full_path, local_path)

    def delete_artifacts(self, artifact_path=None):
        (bucket, dest_path) = self.parse_s3_compliant_uri(self.artifact_uri)
        if artifact_path:
            dest_path = posixpath.join(dest_path, artifact_path)

        dest_path = dest_path.rstrip("/") if dest_path else ""
        s3_client = self._get_s3_client()
        paginator = s3_client.get_paginator("list_objects_v2")
        results = paginator.paginate(Bucket=bucket, Prefix=dest_path)
        for result in results:
            keys = []
            for to_delete_obj in result.get("Contents", []):
                file_path = to_delete_obj.get("Key")
                self._verify_listed_object_contains_artifact_path_prefix(
                    listed_object_path=file_path, artifact_path=dest_path
                )
                keys.append({"Key": file_path})
            if keys:
                s3_client.delete_objects(Bucket=bucket, Delete={"Objects": keys})

    def create_multipart_upload(self, local_file, num_parts=1, artifact_path=None):
        """
        Initiate a multipart upload for efficient large file uploads to S3.

        This method creates a multipart upload session in S3 and generates
        presigned URLs for uploading each part. This is more efficient than
        single-part uploads for large files and provides better error recovery.

        Args:
            local_file: Absolute path to the local file to upload. The file must
                exist and be readable.
            num_parts: Number of parts to split the upload into. Must be between
                1 and 10,000 (S3 limit). More parts allow greater parallelism
                but increase overhead.
            artifact_path: Optional relative path within the S3 bucket where the
                artifact should be stored. If None, the file is stored in the root
                of the configured S3 path.

        Returns:
            CreateMultipartUploadResponse containing:
            - credentials: List of MultipartUploadCredential objects with presigned URLs
            - upload_id: S3 upload ID for tracking this multipart upload
        """
        (bucket, dest_path) = self.parse_s3_compliant_uri(self.artifact_uri)
        if artifact_path:
            dest_path = posixpath.join(dest_path, artifact_path)
        dest_path = posixpath.join(dest_path, os.path.basename(local_file))
        s3_client = self._get_s3_client()
        create_response = s3_client.create_multipart_upload(
            Bucket=bucket,
            Key=dest_path,
        )
        upload_id = create_response["UploadId"]
        credentials = []
        for i in range(1, num_parts + 1):  # part number must be in [1, 10000]
            url = s3_client.generate_presigned_url(
                "upload_part",
                Params={
                    "Bucket": bucket,
                    "Key": dest_path,
                    "PartNumber": i,
                    "UploadId": upload_id,
                },
            )
            credentials.append(
                MultipartUploadCredential(
                    url=url,
                    part_number=i,
                    headers={},
                )
            )
        return CreateMultipartUploadResponse(
            credentials=credentials,
            upload_id=upload_id,
        )

    def complete_multipart_upload(self, local_file, upload_id, parts=None, artifact_path=None):
        """
        Complete a multipart upload by combining all parts into a single S3 object.

        This method should be called after all parts have been successfully uploaded
        using the presigned URLs from create_multipart_upload. It tells S3 to combine
        all the parts into the final object.

        Args:
            local_file: Absolute path to the local file that was uploaded. Must match
                the local_file used in create_multipart_upload.
            upload_id: The S3 upload ID returned by create_multipart_upload.
            parts: List of MultipartUploadPart objects containing metadata for each
                successfully uploaded part. Must include part_number and etag for each part.
                Parts must be provided in order (part 1, part 2, etc.).
            artifact_path: Optional relative path where the artifact should be stored.
                Must match the artifact_path used in create_multipart_upload.
        """
        (bucket, dest_path) = self.parse_s3_compliant_uri(self.artifact_uri)
        if artifact_path:
            dest_path = posixpath.join(dest_path, artifact_path)
        dest_path = posixpath.join(dest_path, os.path.basename(local_file))
        parts = [{"PartNumber": part.part_number, "ETag": part.etag} for part in parts]
        s3_client = self._get_s3_client()
        s3_client.complete_multipart_upload(
            Bucket=bucket, Key=dest_path, UploadId=upload_id, MultipartUpload={"Parts": parts}
        )

    def abort_multipart_upload(self, local_file, upload_id, artifact_path=None):
        """
        Abort a multipart upload and clean up any uploaded parts.

        This method should be called if a multipart upload fails or is cancelled.
        It cleans up any parts that were successfully uploaded and cancels the
        multipart upload session in S3.

        Args:
            local_file: Absolute path to the local file that was being uploaded.
                Must match the local_file used in create_multipart_upload.
            upload_id: The S3 upload ID returned by create_multipart_upload.
            artifact_path: Optional relative path where the artifact would have been stored.
                Must match the artifact_path used in create_multipart_upload.
        """
        (bucket, dest_path) = self.parse_s3_compliant_uri(self.artifact_uri)
        if artifact_path:
            dest_path = posixpath.join(dest_path, artifact_path)
        dest_path = posixpath.join(dest_path, os.path.basename(local_file))
        s3_client = self._get_s3_client()
        s3_client.abort_multipart_upload(
            Bucket=bucket,
            Key=dest_path,
            UploadId=upload_id,
        )



"""
The ``mlflow.prophet`` module provides an API for logging and loading Prophet models.
This module exports univariate Prophet models in the following flavors:

Prophet (native) format
    This is the main flavor that can be accessed with Prophet APIs.
:py:mod:`mlflow.pyfunc`
    Produced for use by generic pyfunc-based deployment tools and for batch auditing
    of historical forecasts.

.. _Prophet:
    https://facebook.github.io/prophet/docs/quick_start.html#python-api
"""

import json
import logging
import os
from typing import Any

import yaml

import mlflow
from mlflow import pyfunc
from mlflow.models import Model, ModelInputExample, ModelSignature
from mlflow.models.model import MLMODEL_FILE_NAME
from mlflow.models.signature import _infer_signature_from_input_example
from mlflow.models.utils import _save_example
from mlflow.tracking._model_registry import DEFAULT_AWAIT_MAX_SLEEP_SECONDS
from mlflow.tracking.artifact_utils import _download_artifact_from_uri
from mlflow.utils.docstring_utils import LOG_MODEL_PARAM_DOCS, format_docstring
from mlflow.utils.environment import (
    _CONDA_ENV_FILE_NAME,
    _CONSTRAINTS_FILE_NAME,
    _PYTHON_ENV_FILE_NAME,
    _REQUIREMENTS_FILE_NAME,
    _mlflow_conda_env,
    _process_conda_env,
    _process_pip_requirements,
    _PythonEnv,
    _validate_env_arguments,
)
from mlflow.utils.file_utils import get_total_file_size, write_to
from mlflow.utils.model_utils import (
    _add_code_from_conf_to_system_path,
    _get_flavor_configuration,
    _validate_and_copy_code_paths,
    _validate_and_prepare_target_save_path,
)
from mlflow.utils.requirements_utils import _get_pinned_requirement

FLAVOR_NAME = "prophet"
_MODEL_BINARY_KEY = "data"
_MODEL_BINARY_FILE_NAME = "model.pr"
_MODEL_TYPE_KEY = "model_type"

_logger = logging.getLogger(__name__)


def get_default_pip_requirements():
    """
    Returns:
        A list of default pip requirements for MLflow Models produced by this flavor.
        Calls to :func:`save_model()` and :func:`log_model()` produce a pip environment
        that, at a minimum, contains these requirements.
    """
    # Note: Prophet's whl build process will fail due to missing dependencies, defaulting
    # to setup.py installation process.
    # If a pystan installation error occurs, ensure gcc>=8 is installed in your environment.
    # See: https://gcc.gnu.org/install/
    return [_get_pinned_requirement("prophet")]


def get_default_conda_env():
    """
    Returns:
        The default Conda environment for MLflow Models produced by calls to
        :func:`save_model()` and :func:`log_model()`.
    """
    return _mlflow_conda_env(additional_pip_deps=get_default_pip_requirements())


@format_docstring(LOG_MODEL_PARAM_DOCS.format(package_name=FLAVOR_NAME))
def save_model(
    pr_model,
    path,
    conda_env=None,
    code_paths=None,
    mlflow_model=None,
    signature: ModelSignature = None,
    input_example: ModelInputExample = None,
    pip_requirements=None,
    extra_pip_requirements=None,
    metadata=None,
):
    """
    Save a Prophet model to a path on the local file system.

    Args:
        pr_model: Prophet model (an instance of Prophet() forecaster that has been fit
            on a temporal series.
        path: Local path where the serialized model (as JSON) is to be saved.
        conda_env: {{ conda_env }}
        code_paths: {{ code_paths }}
        mlflow_model: :py:mod:`mlflow.models.Model` this flavor is being added to.
        signature: an instance of the :py:class:`ModelSignature <mlflow.models.ModelSignature>`
            class that describes the model's inputs and outputs. If not specified but an
            ``input_example`` is supplied, a signature will be automatically inferred
            based on the supplied input example and model. To disable automatic signature
            inference when providing an input example, set ``signature`` to ``False``.
            To manually infer a model signature, call
            :py:func:`infer_signature() <mlflow.models.infer_signature>` on datasets
            with valid model inputs, such as a training dataset with the target column
            omitted, and valid model outputs, like model predictions made on the training
            dataset, for example:

            .. code-block:: python

                from mlflow.models import infer_signature

                model = Prophet().fit(df)
                train = model.history
                predictions = model.predict(model.make_future_dataframe(30))
                signature = infer_signature(train, predictions)
        input_example: {{ input_example }}
        pip_requirements: {{ pip_requirements }}
        extra_pip_requirements: {{ extra_pip_requirements }}
        metadata: {{ metadata }}
    """
    import prophet

    _validate_env_arguments(conda_env, pip_requirements, extra_pip_requirements)

    path = os.path.abspath(path)
    _validate_and_prepare_target_save_path(path)
    code_dir_subpath = _validate_and_copy_code_paths(code_paths, path)

    if mlflow_model is None:
        mlflow_model = Model()
    saved_example = _save_example(mlflow_model, input_example, path)

    if signature is None and saved_example is not None:
        wrapped_model = _ProphetModelWrapper(pr_model)
        signature = _infer_signature_from_input_example(saved_example, wrapped_model)
    elif signature is False:
        signature = None

    if signature is not None:
        mlflow_model.signature = signature
    if metadata is not None:
        mlflow_model.metadata = metadata

    model_data_path = os.path.join(path, _MODEL_BINARY_FILE_NAME)
    _save_model(pr_model, model_data_path)

    model_bin_kwargs = {_MODEL_BINARY_KEY: _MODEL_BINARY_FILE_NAME}
    pyfunc.add_to_model(
        mlflow_model,
        loader_module="mlflow.prophet",
        conda_env=_CONDA_ENV_FILE_NAME,
        python_env=_PYTHON_ENV_FILE_NAME,
        code=code_dir_subpath,
        **model_bin_kwargs,
    )
    flavor_conf = {
        _MODEL_TYPE_KEY: pr_model.__class__.__name__,
        **model_bin_kwargs,
    }
    mlflow_model.add_flavor(
        FLAVOR_NAME,
        prophet_version=prophet.__version__,
        code=code_dir_subpath,
        **flavor_conf,
    )
    if size := get_total_file_size(path):
        mlflow_model.model_size_bytes = size
    mlflow_model.save(os.path.join(path, MLMODEL_FILE_NAME))

    if conda_env is None:
        default_reqs = None
        if pip_requirements is None:
            # cannot use inferred requirements due to prophet's build process
            # as the package installation of pystan requires Cython to be present
            # in the path. Prophet's installation itself requires imports of
            # existing libraries, preventing the execution of a batched pip install
            # and instead using a a strictly defined list of dependencies.
            # NOTE: if Prophet .whl build architecture is changed, this should be
            # modified to a standard inferred approach.
            default_reqs = get_default_pip_requirements()
        conda_env, pip_requirements, pip_constraints = _process_pip_requirements(
            default_reqs,
            pip_requirements,
            extra_pip_requirements,
        )
    else:
        conda_env, pip_requirements, pip_constraints = _process_conda_env(conda_env)

    with open(os.path.join(path, _CONDA_ENV_FILE_NAME), "w") as f:
        yaml.safe_dump(conda_env, stream=f, default_flow_style=False)

    if pip_constraints:
        write_to(os.path.join(path, _CONSTRAINTS_FILE_NAME), "\n".join(pip_constraints))

    write_to(os.path.join(path, _REQUIREMENTS_FILE_NAME), "\n".join(pip_requirements))

    _PythonEnv.current().to_yaml(os.path.join(path, _PYTHON_ENV_FILE_NAME))


@format_docstring(LOG_MODEL_PARAM_DOCS.format(package_name=FLAVOR_NAME))
def log_model(
    pr_model,
    artifact_path: str | None = None,
    conda_env=None,
    code_paths=None,
    registered_model_name=None,
    signature: ModelSignature = None,
    input_example: ModelInputExample = None,
    await_registration_for=DEFAULT_AWAIT_MAX_SLEEP_SECONDS,
    pip_requirements=None,
    extra_pip_requirements=None,
    metadata=None,
    name: str | None = None,
    params: dict[str, Any] | None = None,
    tags: dict[str, Any] | None = None,
    model_type: str | None = None,
    step: int = 0,
    model_id: str | None = None,
):
    """
    Logs a Prophet model as an MLflow artifact for the current run.

    Args:
        pr_model: Prophet model to be saved.
        artifact_path: Deprecated. Use `name` instead.
        conda_env: {{ conda_env }}
        code_paths: {{ code_paths }}
        registered_model_name: If given, create a model
            version under ``registered_model_name``, also creating a
            registered model if one with the given name does not exist.
        signature: An instance of the :py:class:`ModelSignature <mlflow.models.ModelSignature>`
            class that describes the model's inputs and outputs. If not specified but an
            ``input_example`` is supplied, a signature will be automatically inferred
            based on the supplied input example and model. To disable automatic signature
            inference when providing an input example, set ``signature`` to ``False``.
            To manually infer a model signature, call
            :py:func:`infer_signature() <mlflow.models.infer_signature>` on datasets
            with valid model inputs, such as a training dataset with the target column
            omitted, and valid model outputs, like model predictions made on the training
            dataset, for example:

            .. code-block:: python

                from mlflow.models import infer_signature

                model = Prophet().fit(df)
                train = model.history
                predictions = model.predict(model.make_future_dataframe(30))
                signature = infer_signature(train, predictions)

        input_example: {{ input_example }}
        await_registration_for: Number of seconds to wait for the model version
            to finish being created and is in ``READY`` status.
            By default, the function waits for five minutes.
            Specify 0 or None to skip waiting.
        pip_requirements: {{ pip_requirements }}
        extra_pip_requirements: {{ extra_pip_requirements }}
        metadata: {{ metadata }}
        name: {{ name }}
        params: {{ params }}
        tags: {{ tags }}
        model_type: {{ model_type }}
        step: {{ step }}
        model_id: {{ model_id }}

    Returns:
        A :py:class:`ModelInfo <mlflow.models.model.ModelInfo>` instance that contains the
        metadata of the logged model.
    """
    return Model.log(
        artifact_path=artifact_path,
        name=name,
        flavor=mlflow.prophet,
        registered_model_name=registered_model_name,
        pr_model=pr_model,
        conda_env=conda_env,
        code_paths=code_paths,
        signature=signature,
        input_example=input_example,
        await_registration_for=await_registration_for,
        pip_requirements=pip_requirements,
        extra_pip_requirements=extra_pip_requirements,
        metadata=metadata,
        params=params,
        tags=tags,
        model_type=model_type,
        step=step,
        model_id=model_id,
    )


def _save_model(model, path):
    from prophet.serialize import model_to_json

    model_ser = model_to_json(model)
    with open(path, "w") as f:
        json.dump(model_ser, f)


def _load_model(path):
    from prophet.serialize import model_from_json

    with open(path) as f:
        model = json.load(f)
    return model_from_json(model)


def _load_pyfunc(path):
    """
    Loads PyFunc implementation for Prophet. Called by ``pyfunc.load_model``.

    Args:
        path: Local filesystem path to the MLflow Model with the ``prophet`` flavor.
    """
    return _ProphetModelWrapper(_load_model(path))


def load_model(model_uri, dst_path=None):
    """
    Load a Prophet model from a local file or a run.

    Args:
        model_uri: The location, in URI format, of the MLflow model. For example:

            - ``/Users/me/path/to/local/model``
            - ``relative/path/to/local/model``
            - ``s3://my_bucket/path/to/model``
            - ``runs:/<mlflow_run_id>/run-relative/path/to/model``

            For more information about supported URI schemes, see
            `Referencing Artifacts <https://www.mlflow.org/docs/latest/tracking.html#
            artifact-locations>`_.
        dst_path: The local filesystem path to which to download the model artifact.
            This directory must already exist. If unspecified, a local output
            path will be created.

    Returns:
        A Prophet model instance
    """
    local_model_path = _download_artifact_from_uri(artifact_uri=model_uri, output_path=dst_path)
    flavor_conf = _get_flavor_configuration(model_path=local_model_path, flavor_name=FLAVOR_NAME)
    _add_code_from_conf_to_system_path(local_model_path, flavor_conf)
    pr_model_path = os.path.join(
        local_model_path, flavor_conf.get(_MODEL_BINARY_KEY, _MODEL_BINARY_FILE_NAME)
    )

    return _load_model(pr_model_path)


class _ProphetModelWrapper:
    def __init__(self, pr_model):
        self.pr_model = pr_model

    def get_raw_model(self):
        """
        Returns the underlying model.
        """
        return self.pr_model

    def predict(self, dataframe, params: dict[str, Any] | None = None):
        """
        Args:
            dataframe: Model input data.
            params: Additional parameters to pass to the model for inference.

        Returns:
            Model predictions.
        """
        return self.pr_model.predict(dataframe)



"""MLflow tracing integration for Claude Code interactions."""

import json
import logging
import os
import sys
from datetime import datetime
from pathlib import Path
from typing import Any

import dateutil.parser

# ============================================================================
# CONSTANTS
# ============================================================================

# Used multiple times across the module
NANOSECONDS_PER_MS = 1e6
NANOSECONDS_PER_S = 1e9
MAX_PREVIEW_LENGTH = 1000

MESSAGE_TYPE_USER = "user"
MESSAGE_TYPE_ASSISTANT = "assistant"
CONTENT_TYPE_TEXT = "text"
CONTENT_TYPE_TOOL_USE = "tool_use"
CONTENT_TYPE_TOOL_RESULT = "tool_result"
MESSAGE_FIELD_CONTENT = "content"
MESSAGE_FIELD_TYPE = "type"
MESSAGE_FIELD_MESSAGE = "message"
MESSAGE_FIELD_TIMESTAMP = "timestamp"


# ============================================================================
# LOGGING AND SETUP
# ============================================================================


def setup_logging() -> logging.Logger:
    """Set up logging directory and return configured logger.

    Creates .claude/mlflow directory structure and configures file-based logging
    with INFO level. Prevents log propagation to avoid duplicate messages.
    """
    # Create logging directory structure
    log_dir = Path(os.getcwd()) / ".claude" / "mlflow"
    log_dir.mkdir(parents=True, exist_ok=True)

    logger = logging.getLogger(__name__)
    logger.handlers.clear()  # Remove any existing handlers

    # Configure file handler with timestamp formatting
    log_file = log_dir / "claude_tracing.log"
    file_handler = logging.FileHandler(log_file)
    file_handler.setFormatter(
        logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    )
    logger.addHandler(file_handler)
    logger.setLevel(logging.WARNING)
    logger.propagate = False  # Prevent duplicate log messages

    return logger


_MODULE_LOGGER = setup_logging()


def get_logger() -> logging.Logger:
    """Get the configured module logger."""
    return _MODULE_LOGGER


def setup_mlflow() -> None:
    """Configure MLflow tracking URI and experiment."""
    if not is_tracing_enabled():
        return

    import mlflow
    from mlflow.claude_code.config import get_env_var
    from mlflow.environment_variables import (
        MLFLOW_EXPERIMENT_ID,
        MLFLOW_EXPERIMENT_NAME,
        MLFLOW_TRACKING_URI,
    )

    # Get tracking URI from environment/settings
    mlflow.set_tracking_uri(get_env_var(MLFLOW_TRACKING_URI.name))

    # Set experiment if specified via environment variables
    experiment_id = get_env_var(MLFLOW_EXPERIMENT_ID.name)
    experiment_name = get_env_var(MLFLOW_EXPERIMENT_NAME.name)

    try:
        if experiment_id:
            mlflow.set_experiment(experiment_id=experiment_id)
        elif experiment_name:
            mlflow.set_experiment(experiment_name)
    except Exception as e:
        get_logger().warning("Failed to set experiment: %s", e)


def is_tracing_enabled() -> bool:
    """Check if MLflow Claude tracing is enabled via environment variable."""
    try:
        import mlflow  # noqa: F401
    except ImportError as e:
        get_logger().error("MLflow not available: %s", e)
        return False

    from mlflow.claude_code.config import MLFLOW_TRACING_ENABLED, get_env_var

    return get_env_var(MLFLOW_TRACING_ENABLED).lower() in ("true", "1", "yes")


# ============================================================================
# INPUT/OUTPUT UTILITIES
# ============================================================================


def read_hook_input() -> dict[str, Any]:
    """Read JSON input from stdin for Claude Code hook processing."""
    try:
        input_data = sys.stdin.read()
        return json.loads(input_data)
    except json.JSONDecodeError as e:
        get_logger().error("Failed to parse input JSON: %s", e)
        return {}


def read_transcript(transcript_path: str) -> list[dict[str, Any]]:
    """Read and parse a Claude Code conversation transcript from JSONL file."""
    try:
        with open(transcript_path, encoding="utf-8") as f:
            lines = f.readlines()
            return [json.loads(line) for line in lines if line.strip()]
    except Exception as e:
        get_logger().error("Failed to read transcript %s: %s", transcript_path, e)
        return []


def output_hook_response(error: str | None = None, **kwargs) -> None:
    """Output hook response JSON to stdout for Claude Code hook protocol."""
    if error is not None:
        response = {"continue": False, "stopReason": error, **kwargs}
    else:
        response = {"continue": True, **kwargs}

    print(json.dumps(response))  # noqa: T201


# ============================================================================
# TIMESTAMP AND CONTENT PARSING UTILITIES
# ============================================================================


def parse_timestamp_to_ns(timestamp: str | int | float | None) -> int | None:
    """Convert various timestamp formats to nanoseconds since Unix epoch.

    Args:
        timestamp: Can be ISO string, Unix timestamp (seconds/ms), or nanoseconds

    Returns:
        Nanoseconds since Unix epoch, or None if parsing fails
    """
    if not timestamp:
        return None

    if isinstance(timestamp, str):
        try:
            dt = dateutil.parser.parse(timestamp)
            return int(dt.timestamp() * NANOSECONDS_PER_S)
        except Exception:
            get_logger().warning("Could not parse timestamp: %s", timestamp)
            return None
    if isinstance(timestamp, (int, float)):
        if timestamp < 1e10:
            return int(timestamp * NANOSECONDS_PER_S)
        if timestamp < 1e13:
            return int(timestamp * NANOSECONDS_PER_MS)
        return int(timestamp)

    return None


def extract_text_content(content: str | list[dict[str, Any]] | Any) -> str:
    """Extract text content from Claude message content (handles both string and list formats).

    Args:
        content: Either a string or list of content parts from Claude API

    Returns:
        Extracted text content, empty string if none found
    """
    if isinstance(content, list):
        text_parts = []
        for part in content:
            if isinstance(part, dict) and part.get(MESSAGE_FIELD_TYPE) == CONTENT_TYPE_TEXT:
                text_parts.append(part.get(CONTENT_TYPE_TEXT, ""))
        return "\n".join(text_parts)
    if isinstance(content, str):
        return content
    return str(content)


def find_last_user_message_index(transcript: list[dict[str, Any]]) -> int | None:
    """Find the index of the last actual user message (ignoring tool results and empty messages).

    Args:
        transcript: List of conversation entries from Claude Code transcript

    Returns:
        Index of last user message, or None if not found
    """
    for i in range(len(transcript) - 1, -1, -1):
        entry = transcript[i]
        if entry.get(MESSAGE_FIELD_TYPE) == MESSAGE_TYPE_USER and not entry.get("toolUseResult"):
            msg = entry.get(MESSAGE_FIELD_MESSAGE, {})
            content = msg.get(MESSAGE_FIELD_CONTENT, "")

            if isinstance(content, list) and len(content) > 0:
                if (
                    isinstance(content[0], dict)
                    and content[0].get(MESSAGE_FIELD_TYPE) == CONTENT_TYPE_TOOL_RESULT
                ):
                    continue

            if isinstance(content, str) and "<local-command-stdout>" in content:
                continue

            if not content or (isinstance(content, str) and content.strip() == ""):
                continue

            return i
    return None


# ============================================================================
# TRANSCRIPT PROCESSING HELPERS
# ============================================================================


def _get_next_timestamp_ns(transcript: list[dict[str, Any]], current_idx: int) -> int | None:
    """Get the timestamp of the next entry for duration calculation."""
    for i in range(current_idx + 1, len(transcript)):
        timestamp = transcript[i].get(MESSAGE_FIELD_TIMESTAMP)
        if timestamp:
            return parse_timestamp_to_ns(timestamp)
    return None


def _extract_content_and_tools(content: list[dict[str, Any]]) -> tuple[str, list[dict[str, Any]]]:
    """Extract text content and tool uses from assistant response content."""
    text_content = ""
    tool_uses = []

    if isinstance(content, list):
        for part in content:
            if isinstance(part, dict):
                if part.get(MESSAGE_FIELD_TYPE) == CONTENT_TYPE_TEXT:
                    text_content += part.get(CONTENT_TYPE_TEXT, "")
                elif part.get(MESSAGE_FIELD_TYPE) == CONTENT_TYPE_TOOL_USE:
                    tool_uses.append(part)

    return text_content, tool_uses


def _find_tool_results(transcript: list[dict[str, Any]], start_idx: int) -> dict[str, Any]:
    """Find tool results following the current assistant response.

    Returns a mapping from tool_use_id to tool result content.
    """
    tool_results = {}

    # Look for tool results in subsequent entries
    for i in range(start_idx + 1, len(transcript)):
        entry = transcript[i]
        if entry.get(MESSAGE_FIELD_TYPE) != MESSAGE_TYPE_USER:
            continue

        msg = entry.get(MESSAGE_FIELD_MESSAGE, {})
        content = msg.get(MESSAGE_FIELD_CONTENT, [])

        if isinstance(content, list):
            for part in content:
                if (
                    isinstance(part, dict)
                    and part.get(MESSAGE_FIELD_TYPE) == CONTENT_TYPE_TOOL_RESULT
                ):
                    tool_use_id = part.get("tool_use_id")
                    result_content = part.get("content", "")
                    if tool_use_id:
                        tool_results[tool_use_id] = result_content

        # Stop looking once we hit the next assistant response
        if entry.get(MESSAGE_FIELD_TYPE) == MESSAGE_TYPE_ASSISTANT:
            break

    return tool_results


def _reconstruct_conversation_messages(
    transcript: list[dict[str, Any]], end_idx: int
) -> list[dict[str, Any]]:
    """Reconstruct conversation messages in OpenAI format for LLM span inputs.

    This function builds the message array that represents what was sent to the LLM.
    It processes the transcript up to (but not including) end_idx to build the context.

    Args:
        transcript: List of conversation entries from Claude Code transcript
        end_idx: Index to stop at (exclusive) - typically the current assistant response

    Returns:
        List of messages in format [{"role": "system"|"user"|"assistant"|"tool", "content": "..."}]
    """
    messages = []

    for i in range(end_idx):
        entry = transcript[i]
        entry_type = entry.get(MESSAGE_FIELD_TYPE)
        msg = entry.get(MESSAGE_FIELD_MESSAGE, {})

        # Check for system role explicitly
        if msg.get("role") == "system":
            _process_system_entry(msg, messages)
        elif entry_type == MESSAGE_TYPE_USER:
            _process_user_entry(msg, messages)
        elif entry_type == MESSAGE_TYPE_ASSISTANT:
            _process_assistant_entry(msg, messages)

    return messages


def _process_system_entry(msg: dict[str, Any], messages: list[dict[str, Any]]) -> None:
    """Process a system entry from the transcript.

    Args:
        msg: The message object from the entry
        messages: The messages list to append to
    """
    if content := msg.get(MESSAGE_FIELD_CONTENT):
        text_content = extract_text_content(content)
        if text_content.strip():
            messages.append({"role": "system", "content": text_content})


def _process_user_entry(msg: dict[str, Any], messages: list[dict[str, Any]]) -> None:
    """Process a user entry from the transcript and add appropriate messages.

    User entries can contain:
    - Regular user messages (text)
    - Tool results from previous tool calls

    Args:
        msg: The message object from the entry
        messages: The messages list to append to
    """
    content = msg.get(MESSAGE_FIELD_CONTENT, [])

    # Handle list content (typical structure)
    if isinstance(content, list):
        # Use a buffer to preserve original message ordering
        message_buffer = []
        current_text_parts = []

        for part in content:
            if not isinstance(part, dict):
                continue

            part_type = part.get(MESSAGE_FIELD_TYPE)

            if part_type == CONTENT_TYPE_TOOL_RESULT:
                # If we have accumulated text, add it as a user message first
                if current_text_parts:
                    combined_text = "\n".join(current_text_parts).strip()
                    if combined_text:
                        message_buffer.append({"role": "user", "content": combined_text})
                    current_text_parts = []

                # Extract tool result information
                tool_id = part.get("tool_use_id")
                result_content = part.get("content")

                # Add tool results with proper "tool" role
                if result_content:
                    tool_msg = {
                        "role": "tool",
                        "content": result_content,
                    }
                    if tool_id:
                        tool_msg["tool_use_id"] = tool_id
                    message_buffer.append(tool_msg)

            elif part_type == CONTENT_TYPE_TEXT:
                # Accumulate text content
                text = part.get(CONTENT_TYPE_TEXT)
                if text:
                    current_text_parts.append(text)

        # Add any remaining text content as user message
        if current_text_parts:
            combined_text = "\n".join(current_text_parts).strip()
            if combined_text:
                message_buffer.append({"role": "user", "content": combined_text})

        # Add all messages in order to preserve sequence
        messages.extend(message_buffer)

    # Handle string content (simpler format)
    elif isinstance(content, str) and content.strip():
        messages.append({"role": "user", "content": content})


def _process_assistant_entry(msg: dict[str, Any], messages: list[dict[str, Any]]) -> None:
    """Process an assistant entry from the transcript and add to messages.

    Assistant entries represent previous LLM responses that are part of the conversation context.

    Args:
        msg: The message object from the entry
        messages: The messages list to append to
    """
    if content := msg.get(MESSAGE_FIELD_CONTENT):
        text_content = extract_text_content(content)
        if text_content.strip():
            messages.append({"role": "assistant", "content": text_content})


def _create_llm_and_tool_spans(
    client, trace, transcript: list[dict[str, Any]], start_idx: int
) -> None:
    """Create LLM and tool spans for assistant responses with proper timing."""
    from mlflow.entities import SpanType

    llm_call_num = 0
    for i in range(start_idx, len(transcript)):
        entry = transcript[i]
        if entry.get(MESSAGE_FIELD_TYPE) != MESSAGE_TYPE_ASSISTANT:
            continue

        timestamp_ns = parse_timestamp_to_ns(entry.get(MESSAGE_FIELD_TIMESTAMP))
        next_timestamp_ns = _get_next_timestamp_ns(transcript, i)

        # Calculate duration based on next timestamp or use default
        if next_timestamp_ns:
            duration_ns = next_timestamp_ns - timestamp_ns
        else:
            duration_ns = int(1000 * NANOSECONDS_PER_MS)  # 1 second default

        msg = entry.get(MESSAGE_FIELD_MESSAGE, {})
        content = msg.get(MESSAGE_FIELD_CONTENT, [])
        usage = msg.get("usage", {})

        # First check if we have meaningful content to create a span for
        text_content, tool_uses = _extract_content_and_tools(content)

        # Only create LLM span if there's text content (no tools)
        llm_span = None
        if text_content and text_content.strip() and not tool_uses:
            llm_call_num += 1
            conversation_messages = _reconstruct_conversation_messages(transcript, i)

            llm_span = client.start_span(
                name=f"llm_call_{llm_call_num}",
                trace_id=trace.trace_id,
                parent_id=trace.span_id,
                span_type=SpanType.LLM,
                start_time_ns=timestamp_ns,
                inputs={
                    "model": msg.get("model", "unknown"),
                    "messages": conversation_messages,
                },
                attributes={
                    "model": msg.get("model", "unknown"),
                    "input_tokens": usage.get("input_tokens", 0),
                    "output_tokens": usage.get("output_tokens", 0),
                },
            )

            client.end_span(
                trace_id=llm_span.trace_id,
                span_id=llm_span.span_id,
                outputs={"response": text_content},
                end_time_ns=timestamp_ns + duration_ns,
            )

        # Create tool spans with proportional timing and actual results
        if tool_uses:
            tool_results = _find_tool_results(transcript, i)
            tool_duration_ns = duration_ns // len(tool_uses)

            for idx, tool_use in enumerate(tool_uses):
                tool_start_ns = timestamp_ns + (idx * tool_duration_ns)
                tool_use_id = tool_use.get("id", "")
                tool_result = tool_results.get(tool_use_id, "No result found")

                tool_span = client.start_span(
                    name=f"tool_{tool_use.get('name', 'unknown')}",
                    trace_id=trace.trace_id,
                    parent_id=trace.span_id,
                    span_type=SpanType.TOOL,
                    start_time_ns=tool_start_ns,
                    inputs=tool_use.get("input", {}),
                    attributes={
                        "tool_name": tool_use.get("name", "unknown"),
                        "tool_id": tool_use_id,
                    },
                )

                client.end_span(
                    trace_id=tool_span.trace_id,
                    span_id=tool_span.span_id,
                    outputs={"result": tool_result},
                    end_time_ns=tool_start_ns + tool_duration_ns,
                )


def find_final_assistant_response(transcript: list[dict[str, Any]], start_idx: int) -> str | None:
    """Find the final text response from the assistant for trace preview.

    Args:
        transcript: List of conversation entries from Claude Code transcript
        start_idx: Index to start searching from (typically after last user message)

    Returns:
        Final assistant response text or None
    """
    final_response = None

    for i in range(start_idx, len(transcript)):
        entry = transcript[i]
        if entry.get(MESSAGE_FIELD_TYPE) != MESSAGE_TYPE_ASSISTANT:
            continue

        msg = entry.get(MESSAGE_FIELD_MESSAGE, {})
        content = msg.get(MESSAGE_FIELD_CONTENT, [])

        if isinstance(content, list):
            for part in content:
                if isinstance(part, dict) and part.get(MESSAGE_FIELD_TYPE) == CONTENT_TYPE_TEXT:
                    text = part.get(CONTENT_TYPE_TEXT, "")
                    if text.strip():
                        final_response = text

    return final_response


# ============================================================================
# MAIN TRANSCRIPT PROCESSING
# ============================================================================


def process_transcript(transcript_path: str, session_id: str | None = None) -> Any | None:
    """Process a Claude conversation transcript and create an MLflow trace with spans.

    Args:
        transcript_path: Path to the Claude Code transcript.jsonl file
        session_id: Optional session identifier, defaults to timestamp-based ID

    Returns:
        MLflow trace object if successful, None if processing fails
    """
    if not is_tracing_enabled():
        get_logger().info("MLflow Claude tracing is disabled")
        return None

    try:
        import mlflow
        from mlflow import MlflowClient
        from mlflow.tracing.constant import TraceMetadataKey
        from mlflow.tracing.trace_manager import InMemoryTraceManager

        setup_mlflow()
        client = MlflowClient()

        transcript = read_transcript(transcript_path)
        if not transcript:
            get_logger().warning("Empty transcript, skipping")
            return None

        last_user_idx = find_last_user_message_index(transcript)
        if last_user_idx is None:
            get_logger().warning("No user message found in transcript")
            return None

        last_user_entry = transcript[last_user_idx]
        last_user_prompt = last_user_entry.get(MESSAGE_FIELD_MESSAGE, {}).get(
            MESSAGE_FIELD_CONTENT, ""
        )

        if not session_id:
            session_id = f"claude-{datetime.now().strftime('%Y%m%d_%H%M%S')}"

        get_logger().info("Creating MLflow trace for session: %s", session_id)

        conv_start_ns = parse_timestamp_to_ns(last_user_entry.get(MESSAGE_FIELD_TIMESTAMP))

        trace = client.start_trace(
            name="claude_code_conversation",
            inputs={"prompt": extract_text_content(last_user_prompt)},
            start_time_ns=conv_start_ns,
        )

        # Create spans for all assistant responses and tool uses
        _create_llm_and_tool_spans(client, trace, transcript, last_user_idx + 1)

        # Update trace with preview content and end timing
        final_response = find_final_assistant_response(transcript, last_user_idx + 1)
        user_prompt_text = extract_text_content(last_user_prompt)

        # Set trace previews for UI display
        try:
            with InMemoryTraceManager.get_instance().get_trace(trace.trace_id) as in_memory_trace:
                if user_prompt_text:
                    in_memory_trace.info.request_preview = user_prompt_text[:MAX_PREVIEW_LENGTH]
                if final_response:
                    in_memory_trace.info.response_preview = final_response[:MAX_PREVIEW_LENGTH]
                in_memory_trace.info.trace_metadata = {
                    TraceMetadataKey.TRACE_SESSION: session_id,
                    TraceMetadataKey.TRACE_USER: os.environ.get("USER", ""),
                    "mlflow.trace.working_directory": os.getcwd(),
                }
        except Exception as e:
            get_logger().warning("Failed to update trace metadata and previews: %s", e)

        # Calculate end time based on last entry or use default duration
        last_entry = transcript[-1] if transcript else last_user_entry
        conv_end_ns = parse_timestamp_to_ns(last_entry.get(MESSAGE_FIELD_TIMESTAMP))
        if not conv_end_ns or conv_end_ns <= conv_start_ns:
            conv_end_ns = conv_start_ns + int(10 * NANOSECONDS_PER_S)  # 10 second default

        client.end_trace(
            trace_id=trace.trace_id,
            outputs={"response": final_response or "Conversation completed", "status": "completed"},
            end_time_ns=conv_end_ns,
        )

        mlflow.flush_trace_async_logging()
        get_logger().info("Created MLflow trace: %s", trace.trace_id)

        return mlflow.get_trace(trace.trace_id)

    except Exception as e:
        get_logger().error("Error processing transcript: %s", e, exc_info=True)
        return None



import json
import logging
import math
from concurrent.futures import Future, as_completed
from typing import TYPE_CHECKING, Any, Collection

from mlflow.entities import Assessment, Trace, TraceData
from mlflow.entities.assessment import DEFAULT_FEEDBACK_NAME, Feedback
from mlflow.entities.assessment_source import AssessmentSource, AssessmentSourceType
from mlflow.exceptions import MlflowException
from mlflow.genai.evaluation.constant import (
    AgentEvaluationReserverKey,
)
from mlflow.genai.scorers import Scorer
from mlflow.models import EvaluationMetric

try:
    # `pandas` is not required for `mlflow-skinny`.
    import pandas as pd
except ImportError:
    pass

if TYPE_CHECKING:
    from mlflow.entities.evaluation_dataset import EvaluationDataset as EntityEvaluationDataset
    from mlflow.genai.datasets import EvaluationDataset as ManagedEvaluationDataset
    from mlflow.genai.evaluation.entities import EvalResult

    try:
        import pyspark.sql.dataframe

        EvaluationDatasetTypes = (
            pd.DataFrame
            | pyspark.sql.dataframe.DataFrame
            | list[dict]
            | ManagedEvaluationDataset
            | EntityEvaluationDataset
        )
    except ImportError:
        EvaluationDatasetTypes = (
            pd.DataFrame | list[dict] | ManagedEvaluationDataset | EntityEvaluationDataset
        )


_logger = logging.getLogger(__name__)

USER_DEFINED_ASSESSMENT_NAME_KEY = "_user_defined_assessment_name"
PGBAR_FORMAT = (
    "{l_bar}{bar}| {n_fmt}/{total_fmt} [Elapsed: {elapsed}, Remaining: {remaining}] {postfix}"
)


def _convert_eval_set_to_df(data: "EvaluationDatasetTypes") -> "pd.DataFrame":
    """
    Takes in a dataset in the format that `mlflow.genai.evaluate()` expects and
    converts it into a pandas DataFrame.
    """
    if isinstance(data, list):
        # validate that every item in the list is a dict and has inputs as key
        for item in data:
            if not isinstance(item, dict):
                raise MlflowException.invalid_parameter_value(
                    "Every item in the list must be a dictionary."
                )
        df = pd.DataFrame(data)
    elif isinstance(data, pd.DataFrame):
        # Data is already a pd DataFrame, just copy it
        df = data.copy()
    else:
        try:
            from mlflow.utils.spark_utils import get_spark_dataframe_type

            if isinstance(data, get_spark_dataframe_type()):
                df = _deserialize_inputs_and_expectations_column(data.toPandas())
            else:
                raise MlflowException.invalid_parameter_value(
                    "Invalid type for parameter `data`. Expected a list of dictionaries, "
                    f"a pandas DataFrame, or a Spark DataFrame. Got: {type(data)}"
                )
        except ImportError:
            raise ImportError(
                "The `pyspark` package is required to use mlflow.genai.evaluate() "
                "Please install it with `pip install pyspark`."
            )

    if len(df) == 0:
        raise MlflowException.invalid_parameter_value(
            "The dataset is empty. Please provide a non-empty dataset."
        )

    if not any(col in df.columns for col in ("trace", "inputs")):
        raise MlflowException.invalid_parameter_value(
            "Either `inputs` or `trace` column is required in the dataset. Please provide inputs "
            "for every datapoint or provide a trace."
        )

    return df


def _convert_to_eval_set(data: "EvaluationDatasetTypes") -> "pd.DataFrame":
    """
    Takes in a dataset in the multiple format that mlflow.genai.evaluate() expects and converts it
    into standardized Pandas DataFrame.
    The expected schema can be found at:
    https://docs.databricks.com/aws/en/generative-ai/agent-evaluation/evaluation-schema

    NB: The harness secretly support 'expectations' column as well. It accepts a dictionary of
        expectations, which is same as the schema that mlflow.genai.evaluate() expects.
        Therefore, we can simply pass through expectations column.
    """
    column_mapping = {
        "inputs": "request",
        "outputs": "response",
    }

    df = _convert_eval_set_to_df(data)

    return (
        df.rename(columns=column_mapping)
        .pipe(_deserialize_trace_column_if_needed)
        .pipe(_extract_request_response_from_trace)
        .pipe(_extract_expectations_from_trace)
    )


def _deserialize_inputs_and_expectations_column(df: "pd.DataFrame") -> "pd.DataFrame":
    """
    Deserialize the `inputs` and `expectations` string columns from the dataframe.

    When managed datasets are read as Spark DataFrames, the `inputs` and `expectations` columns
    are loaded as string columns of JSON strings. This function deserializes these columns into
    dictionaries expected by mlflow.genai.evaluate().
    """
    target_columns = ["inputs", "expectations"]
    for col in target_columns:
        if col not in df.columns or not isinstance(df[col][0], str):
            continue

        try:
            df[col] = df[col].apply(json.loads)
        except json.JSONDecodeError as e:
            if col == "inputs":
                msg = (
                    "The `inputs` column must be a valid JSON string of field names and values. "
                    "For example, `{'question': 'What is the capital of France?'}`"
                )
            else:
                msg = (
                    "The `expectations` column must be a valid JSON string of assessment names and "
                    "values. For example, `{'expected_facts': ['fact1', 'fact2']}`"
                )
            raise MlflowException.invalid_parameter_value(
                f"Failed to parse `{col}` column. Error: {e}\nHint: {msg}"
            )

    return df


def _deserialize_trace_column_if_needed(df: "pd.DataFrame") -> "pd.DataFrame":
    """
    Deserialize the `trace` column from the dataframe if it is a string.

    Since MLflow 3.2.0, mlflow.search_traces() returns a pandas DataFrame with a `trace`
    column that is a trace json representation rather than the Trace object itself. This
    function deserializes the `trace` column into a Trace object.
    """
    if "trace" in df.columns:
        df["trace"] = df["trace"].apply(lambda t: Trace.from_json(t) if isinstance(t, str) else t)
    return df


def _extract_request_response_from_trace(df: "pd.DataFrame") -> "pd.DataFrame":
    """
    Add `request` and `response`columns from traces if it is not already present.
    """
    if "trace" not in df.columns:
        return df

    def _extract_attribute(trace_data: TraceData, attribute_name: str) -> Any:
        if att := getattr(trace_data, attribute_name, None):
            return json.loads(att)
        return None

    if "request" not in df.columns:
        df["request"] = df["trace"].apply(lambda trace: _extract_attribute(trace.data, "request"))
    if "response" not in df.columns:
        df["response"] = df["trace"].apply(lambda trace: _extract_attribute(trace.data, "response"))
    return df


def _extract_expectations_from_trace(df: "pd.DataFrame") -> "pd.DataFrame":
    """
    Add `expectations` columns to the dataframe from assessments
    stored in the traces, if the "expectations" column is not already present.
    """
    if "trace" not in df.columns:
        return df

    expectations_column = []
    for trace in df["trace"]:
        expectations = {}
        for assessment in trace.info.assessments or []:
            if assessment.expectation is not None:
                expectations[assessment.name] = assessment.expectation.value
        expectations_column.append(expectations)
    # If no trace has assessments, not add the column
    if all(len(expectations) == 0 for expectations in expectations_column):
        return df

    df["expectations"] = expectations_column
    return df


def _convert_scorer_to_legacy_metric(scorer: Scorer) -> EvaluationMetric:
    """
    Takes in a Scorer object and converts it into a legacy MLflow 2.x
    Metric object.
    """
    try:
        from databricks.agents.evals import metric
    except ImportError:
        raise ImportError(
            "The `databricks-agents` package is required to use mlflow.genai.evaluate() "
            "Please install it with `pip install databricks-agents`."
        )

    from mlflow.genai.scorers.builtin_scorers import BuiltInScorer
    from mlflow.types.llm import ChatCompletionRequest

    def eval_fn(
        request_id: str,
        request: ChatCompletionRequest | str,
        response: Any | None,
        expected_response: Any | None,
        trace: Trace | None,
        guidelines: list[str] | dict[str, list[str]] | None,
        expected_facts: list[str] | None,
        expected_retrieved_context: list[dict[str, str]] | None,
        custom_expected: dict[str, Any] | None,
        **kwargs,
    ) -> int | float | bool | str | Assessment | list[Assessment]:
        # Condense all expectations into a single dict
        expectations = {}
        if expected_response is not None:
            expectations[AgentEvaluationReserverKey.EXPECTED_RESPONSE] = expected_response
        if expected_facts is not None:
            expectations[AgentEvaluationReserverKey.EXPECTED_FACTS] = expected_facts
        if expected_retrieved_context is not None:
            expectations[AgentEvaluationReserverKey.EXPECTED_RETRIEVED_CONTEXT] = (
                expected_retrieved_context
            )
        if guidelines is not None:
            expectations[AgentEvaluationReserverKey.GUIDELINES] = guidelines
        if custom_expected is not None:
            expectations.update(custom_expected)

        merged = {
            "inputs": request,
            "outputs": response,
            "expectations": expectations,
            "trace": trace,
        }
        return scorer.run(**merged)

    metric_instance = metric(
        eval_fn=eval_fn,
        name=scorer.name,
    )
    # Add aggregations as an attribute since the metric decorator doesn't accept it
    metric_instance.aggregations = scorer.aggregations
    # Add attribute to indicate if this is a built-in scorer
    metric_instance._is_builtin_scorer = isinstance(scorer, BuiltInScorer)

    return metric_instance


def standardize_scorer_value(scorer_name: str, value: Any) -> list[Feedback]:
    """
    Convert the scorer return value to a list of MLflow Assessment (Feedback) objects.

    Scorer can return:
    - A number, boolean, or string, a list of them.
    - An Feedback object
    - A list of Feedback objects

    All of the above will be converted to a list of Feedback objects.
    """
    # None is a valid metric value, return an empty list
    if value is None:
        return []

    # Primitives are valid metric values
    if isinstance(value, (int, float, bool, str)):
        return [
            Feedback(
                name=scorer_name,
                source=make_code_type_assessment_source(scorer_name),
                value=value,
            )
        ]

    if isinstance(value, Feedback):
        value.name = _get_custom_assessment_name(value, scorer_name)
        return [value]

    if isinstance(value, Collection):
        assessments = []
        for item in value:
            if isinstance(item, Feedback):
                # Scorer returns multiple assessments as a list.
                item.name = _get_custom_assessment_name(item, scorer_name)
                assessments.append(item)
            else:
                # If the item is not assessment, the list represents a single assessment
                # value of list type. Convert it to a Feedback object.
                assessments.append(
                    Feedback(
                        name=scorer_name,
                        source=make_code_type_assessment_source(scorer_name),
                        value=item,
                    )
                )
        return assessments

    raise MlflowException.invalid_parameter_value(
        f"Got unsupported result from scorer '{scorer_name}'. "
        f"Expected the metric value to be a number, or a boolean, or a string, "
        "or an Feedback, or a list of Feedbacks. "
        f"Got {value}.",
    )


def _get_custom_assessment_name(assessment: Feedback, scorer_name: str) -> str:
    """Get the name of the custom assessment. Use assessment name if present and not a builtin judge
    name, otherwise use the scorer name.

    Args:
        assessment: The assessment to get the name for.
        scorer_name: The name of the scprer.
    """
    # If the user didn't provide a name, use the scorer name
    if assessment.name == DEFAULT_FEEDBACK_NAME or (
        assessment.metadata is not None
        and assessment.metadata.get(USER_DEFINED_ASSESSMENT_NAME_KEY) == "false"
    ):
        return scorer_name
    return assessment.name


def make_code_type_assessment_source(scorer_name: str) -> AssessmentSource:
    return AssessmentSource(source_type=AssessmentSourceType.CODE, source_id=scorer_name)


def is_none_or_nan(value: Any) -> bool:
    """
    Checks whether a value is None or NaN.

    NB: This function does not handle pandas.NA.
    """
    # isinstance(value, float) check is needed to ensure that math.isnan is not called on an array.
    return value is None or (isinstance(value, float) and math.isnan(value))


def validate_tags(tags: Any) -> None:
    """
    Validate that tags are in the expected format: dict[str, str].

    Args:
        tags: The tags to validate.

    Raises:
        MlflowException: If tags are not in the correct format.
    """
    if is_none_or_nan(tags):
        return

    if not isinstance(tags, dict):
        raise MlflowException.invalid_parameter_value(
            f"Tags must be a dictionary, got {type(tags).__name__}. "
        )

    errors = []
    for key in tags.keys():
        if not isinstance(key, str):
            errors.append(f"Key {key!r} has type {type(key).__name__}; expected str.")

    if errors:
        raise MlflowException.invalid_parameter_value("Invalid tags:\n  - " + "\n  - ".join(errors))


def complete_eval_futures_with_progress_base(futures: list[Future]) -> list["EvalResult"]:
    """Wraps the as_completed function with a progress bar."""
    futures_as_completed = as_completed(futures)

    try:
        from tqdm.auto import tqdm

        futures_as_completed = tqdm(
            futures_as_completed,
            total=len(futures),
            disable=False,
            desc="Evaluating",
            smoothing=0,  # 0 means using average speed for remaining time estimates
            bar_format=PGBAR_FORMAT,
        )
    except ImportError:
        # If tqdm is not installed, we don't show a progress bar
        pass

    return [future.result() for future in futures_as_completed]



class User:
    def __init__(
        self,
        id_,
        username,
        password_hash,
        is_admin,
        experiment_permissions=None,
        registered_model_permissions=None,
    ):
        self._id = id_
        self._username = username
        self._password_hash = password_hash
        self._is_admin = is_admin
        self._experiment_permissions = experiment_permissions
        self._registered_model_permissions = registered_model_permissions

    @property
    def id(self):
        return self._id

    @property
    def username(self):
        return self._username

    @property
    def password_hash(self):
        return self._password_hash

    @property
    def is_admin(self):
        return self._is_admin

    @is_admin.setter
    def is_admin(self, is_admin):
        self._is_admin = is_admin

    @property
    def experiment_permissions(self):
        return self._experiment_permissions

    @experiment_permissions.setter
    def experiment_permissions(self, experiment_permissions):
        self._experiment_permissions = experiment_permissions

    @property
    def registered_model_permissions(self):
        return self._registered_model_permissions

    @registered_model_permissions.setter
    def registered_model_permissions(self, registered_model_permissions):
        self._registered_model_permissions = registered_model_permissions

    def to_json(self):
        return {
            "id": self.id,
            "username": self.username,
            "is_admin": self.is_admin,
            "experiment_permissions": [p.to_json() for p in self.experiment_permissions],
            "registered_model_permissions": [
                p.to_json() for p in self.registered_model_permissions
            ],
        }

    @classmethod
    def from_json(cls, dictionary):
        return cls(
            id_=dictionary["id"],
            username=dictionary["username"],
            password_hash="REDACTED",
            is_admin=dictionary["is_admin"],
            experiment_permissions=[
                ExperimentPermission.from_json(p) for p in dictionary["experiment_permissions"]
            ],
            registered_model_permissions=[
                RegisteredModelPermission.from_json(p)
                for p in dictionary["registered_model_permissions"]
            ],
        )


class ExperimentPermission:
    def __init__(
        self,
        experiment_id,
        user_id,
        permission,
    ):
        self._experiment_id = experiment_id
        self._user_id = user_id
        self._permission = permission

    @property
    def experiment_id(self):
        return self._experiment_id

    @property
    def user_id(self):
        return self._user_id

    @property
    def permission(self):
        return self._permission

    @permission.setter
    def permission(self, permission):
        self._permission = permission

    def to_json(self):
        return {
            "experiment_id": self.experiment_id,
            "user_id": self.user_id,
            "permission": self.permission,
        }

    @classmethod
    def from_json(cls, dictionary):
        return cls(
            experiment_id=dictionary["experiment_id"],
            user_id=dictionary["user_id"],
            permission=dictionary["permission"],
        )


class RegisteredModelPermission:
    def __init__(
        self,
        name,
        user_id,
        permission,
    ):
        self._name = name
        self._user_id = user_id
        self._permission = permission

    @property
    def name(self):
        return self._name

    @property
    def user_id(self):
        return self._user_id

    @property
    def permission(self):
        return self._permission

    @permission.setter
    def permission(self, permission):
        self._permission = permission

    def to_json(self):
        return {
            "name": self.name,
            "user_id": self.user_id,
            "permission": self.permission,
        }

    @classmethod
    def from_json(cls, dictionary):
        return cls(
            name=dictionary["name"],
            user_id=dictionary["user_id"],
            permission=dictionary["permission"],
        )



import base64

from mlflow.exceptions import MlflowException
from mlflow.protos.databricks_pb2 import INVALID_PARAMETER_VALUE
from mlflow.protos.databricks_uc_registry_messages_pb2 import (
    MODEL_VERSION_OPERATION_READ,
    GenerateTemporaryModelVersionCredentialsRequest,
    GenerateTemporaryModelVersionCredentialsResponse,
    ModelVersionLineageDirection,
    StorageMode,
)
from mlflow.protos.databricks_uc_registry_service_pb2 import UcModelRegistryService
from mlflow.store._unity_catalog.lineage.constants import _DATABRICKS_LINEAGE_ID_HEADER
from mlflow.store.artifact.artifact_repo import ArtifactRepository
from mlflow.store.artifact.databricks_sdk_models_artifact_repo import (
    DatabricksSDKModelsArtifactRepository,
)
from mlflow.store.artifact.presigned_url_artifact_repo import PresignedUrlArtifactRepository
from mlflow.store.artifact.utils.models import (
    get_model_name_and_version,
)
from mlflow.utils._spark_utils import _get_active_spark_session
from mlflow.utils._unity_catalog_utils import (
    emit_model_version_lineage,
    get_artifact_repo_from_storage_info,
    get_full_name_from_sc,
    is_databricks_sdk_models_artifact_repository_enabled,
)
from mlflow.utils.databricks_utils import get_databricks_host_creds
from mlflow.utils.proto_json_utils import message_to_json
from mlflow.utils.rest_utils import (
    _REST_API_PATH_PREFIX,
    call_endpoint,
    extract_api_info_for_service,
)
from mlflow.utils.uri import (
    _DATABRICKS_UNITY_CATALOG_SCHEME,
    get_databricks_profile_uri_from_artifact_uri,
    get_db_info_from_uri,
    is_databricks_unity_catalog_uri,
)

_METHOD_TO_INFO = extract_api_info_for_service(UcModelRegistryService, _REST_API_PATH_PREFIX)


class UnityCatalogModelsArtifactRepository(ArtifactRepository):
    """
    Performs storage operations on artifacts controlled by a Unity Catalog model registry

    Temporary scoped tokens for the appropriate cloud storage locations are fetched from the
    remote backend and used to download model artifacts.

    The artifact_uri is expected to be of the form `models:/<model_name>/<model_version>`

    Note : This artifact repository is meant is to be instantiated by the ModelsArtifactRepository
    when the client is pointing to a Unity Catalog model registry.
    """

    def __init__(self, artifact_uri, registry_uri, tracking_uri: str | None = None):
        if not is_databricks_unity_catalog_uri(registry_uri):
            raise MlflowException(
                message="Attempted to instantiate an artifact repo to access models in the "
                f"Unity Catalog with non-Unity Catalog registry URI '{registry_uri}'. "
                f"Please specify a Unity Catalog registry URI of the "
                f"form '{_DATABRICKS_UNITY_CATALOG_SCHEME}[://profile]', e.g. by calling "
                f"mlflow.set_registry_uri('{_DATABRICKS_UNITY_CATALOG_SCHEME}') if using the "
                f"MLflow Python client",
                error_code=INVALID_PARAMETER_VALUE,
            )
        super().__init__(artifact_uri, tracking_uri, registry_uri)
        from mlflow.tracking.client import MlflowClient

        registry_uri_from_artifact_uri = get_databricks_profile_uri_from_artifact_uri(
            artifact_uri, result_scheme=_DATABRICKS_UNITY_CATALOG_SCHEME
        )
        if registry_uri_from_artifact_uri is not None:
            registry_uri = registry_uri_from_artifact_uri
        _, key_prefix = get_db_info_from_uri(registry_uri)
        if key_prefix is not None:
            raise MlflowException(
                "Remote model registry access via model URIs of the form "
                "'models://<scope>@<prefix>/<model_name>/<version_or_stage>' is unsupported for "
                "models in the Unity Catalog. We recommend that you access the Unity Catalog "
                "from the current Databricks workspace instead."
            )
        self.registry_uri = registry_uri
        self.client = MlflowClient(registry_uri=self.registry_uri)
        try:
            spark = _get_active_spark_session()
        except Exception:
            pass
        model_name, self.model_version = get_model_name_and_version(self.client, artifact_uri)
        self.model_name = get_full_name_from_sc(model_name, spark)

    def _get_blob_storage_path(self):
        return self.client.get_model_version_download_uri(self.model_name, self.model_version)

    def _get_scoped_token(self, lineage_header_info=None):
        extra_headers = {}
        if lineage_header_info:
            header_json = message_to_json(lineage_header_info)
            header_base64 = base64.b64encode(header_json.encode())
            extra_headers[_DATABRICKS_LINEAGE_ID_HEADER] = header_base64

        db_creds = get_databricks_host_creds(self.registry_uri)
        endpoint, method = _METHOD_TO_INFO[GenerateTemporaryModelVersionCredentialsRequest]
        req_body = message_to_json(
            GenerateTemporaryModelVersionCredentialsRequest(
                name=self.model_name,
                version=self.model_version,
                operation=MODEL_VERSION_OPERATION_READ,
            )
        )
        response_proto = GenerateTemporaryModelVersionCredentialsResponse()
        return call_endpoint(
            host_creds=db_creds,
            endpoint=endpoint,
            method=method,
            json_body=req_body,
            response_proto=response_proto,
            extra_headers=extra_headers,
        ).credentials

    def _get_artifact_repo(self, lineage_header_info=None):
        """
        Get underlying ArtifactRepository instance for model version blob
        storage
        """
        host_creds = get_databricks_host_creds(self.registry_uri)
        if is_databricks_sdk_models_artifact_repository_enabled(host_creds):
            entities = lineage_header_info.entities if lineage_header_info else []
            emit_model_version_lineage(
                host_creds,
                self.model_name,
                self.model_version,
                entities,
                ModelVersionLineageDirection.DOWNSTREAM,
            )
            return DatabricksSDKModelsArtifactRepository(self.model_name, self.model_version)
        scoped_token = self._get_scoped_token(lineage_header_info=lineage_header_info)
        if scoped_token.storage_mode == StorageMode.DEFAULT_STORAGE:
            return PresignedUrlArtifactRepository(
                get_databricks_host_creds(self.registry_uri), self.model_name, self.model_version
            )

        blob_storage_path = self._get_blob_storage_path()
        return get_artifact_repo_from_storage_info(
            storage_location=blob_storage_path,
            scoped_token=scoped_token,
            base_credential_refresh_def=self._get_scoped_token,
        )

    def list_artifacts(self, path=None):
        return self._get_artifact_repo().list_artifacts(path=path)

    def download_artifacts(self, artifact_path, dst_path=None, lineage_header_info=None):
        return self._get_artifact_repo(lineage_header_info=lineage_header_info).download_artifacts(
            artifact_path, dst_path
        )

    def log_artifact(self, local_file, artifact_path=None):
        raise MlflowException("This repository does not support logging artifacts.")

    def log_artifacts(self, local_dir, artifact_path=None):
        raise MlflowException("This repository does not support logging artifacts.")

    def delete_artifacts(self, artifact_path=None):
        raise NotImplementedError("This artifact repository does not support deleting artifacts")



from google.protobuf import descriptor_pb2 as _descriptor_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor
OPTIONS_FIELD_NUMBER: _ClassVar[int]
options: _descriptor.FieldDescriptor
MESSAGE_FIELD_NUMBER: _ClassVar[int]
message: _descriptor.FieldDescriptor
FIELD_FIELD_NUMBER: _ClassVar[int]
field: _descriptor.FieldDescriptor

class ScalaPbOptions(_message.Message):
    __slots__ = ("package_name", "flat_package")
    PACKAGE_NAME_FIELD_NUMBER: _ClassVar[int]
    FLAT_PACKAGE_FIELD_NUMBER: _ClassVar[int]
    IMPORT_FIELD_NUMBER: _ClassVar[int]
    package_name: str
    flat_package: bool
    def __init__(self, package_name: _Optional[str] = ..., flat_package: bool = ..., **kwargs) -> None: ...

class MessageOptions(_message.Message):
    __slots__ = ("extends",)
    EXTENDS_FIELD_NUMBER: _ClassVar[int]
    extends: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, extends: _Optional[_Iterable[str]] = ...) -> None: ...

class FieldOptions(_message.Message):
    __slots__ = ("type",)
    TYPE_FIELD_NUMBER: _ClassVar[int]
    type: str
    def __init__(self, type: _Optional[str] = ...) -> None: ...



"""add jobs table

Create Date: 2025-09-11 17:39:31.569736

"""

import time

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "bf29a5ff90ea"
down_revision = "3da73c924c2f"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "jobs",
        sa.Column("id", sa.String(length=36), nullable=False),
        sa.Column("creation_time", sa.BigInteger(), default=lambda: int(time.time() * 1000)),
        sa.Column("function_fullname", sa.String(length=500), nullable=False),
        sa.Column("params", sa.Text(), nullable=False),
        sa.Column("timeout", sa.Float(precision=53), nullable=True),
        sa.Column("status", sa.Integer(), nullable=False),
        sa.Column("result", sa.Text(), nullable=True),
        sa.Column("retry_count", sa.Integer(), default=0),
        sa.PrimaryKeyConstraint("id", name="jobs_pk"),
    )
    with op.batch_alter_table("jobs", schema=None) as batch_op:
        batch_op.create_index(
            "index_jobs_function_status_creation_time",
            ["function_fullname", "status", "creation_time"],
            unique=False,
        )


def downgrade():
    op.drop_table("jobs")



"""MLflow CLI commands for Claude Code integration."""

from pathlib import Path

import click

from mlflow.claude_code.config import get_tracing_status, setup_environment_config
from mlflow.claude_code.hooks import disable_tracing_hooks, setup_hooks_config


@click.group("autolog")
def commands():
    """Commands for autologging with MLflow."""


@commands.command("claude")
@click.argument("directory", default=".", type=click.Path(file_okay=False, dir_okay=True))
@click.option(
    "--tracking-uri", "-u", help="MLflow tracking URI (e.g., 'databricks' or 'file://mlruns')"
)
@click.option("--experiment-id", "-e", help="MLflow experiment ID")
@click.option("--experiment-name", "-n", help="MLflow experiment name")
@click.option("--disable", is_flag=True, help="Disable Claude tracing in the specified directory")
@click.option("--status", is_flag=True, help="Show current tracing status")
def claude(
    directory: str,
    tracking_uri: str | None,
    experiment_id: str | None,
    experiment_name: str | None,
    disable: bool,
    status: bool,
) -> None:
    """Set up Claude Code tracing in a directory.

    This command configures Claude Code hooks to automatically trace conversations
    to MLflow. After setup, use the regular 'claude' command and traces will be
    automatically created.

    DIRECTORY: Directory to set up tracing in (default: current directory)

    Examples:

      # Set up tracing in current directory with local storage
      mlflow autolog claude

      # Set up tracing in a specific project directory
      mlflow autolog claude ~/my-project

      # Set up tracing with Databricks
      mlflow autolog claude -u databricks -e 123456789

      # Set up tracing with custom tracking URI
      mlflow autolog claude -u file://./custom-mlruns

      # Disable tracing in current directory
      mlflow autolog claude --disable
    """
    target_dir = Path(directory).resolve()
    claude_dir = target_dir / ".claude"
    settings_file = claude_dir / "settings.json"

    if status:
        _show_status(target_dir, settings_file)
        return

    if disable:
        _handle_disable(settings_file)
        return

    click.echo(f"Configuring Claude tracing in: {target_dir}")

    # Create .claude directory and set up hooks
    claude_dir.mkdir(parents=True, exist_ok=True)
    setup_hooks_config(settings_file)
    click.echo("✅ Claude Code hooks configured")

    # Set up environment variables
    setup_environment_config(settings_file, tracking_uri, experiment_id, experiment_name)

    # Show final status
    _show_setup_status(target_dir, tracking_uri, experiment_id, experiment_name)


def _handle_disable(settings_file: Path) -> None:
    """Handle disable command."""
    if disable_tracing_hooks(settings_file):
        click.echo("✅ Claude tracing disabled")
    else:
        click.echo("❌ No Claude configuration found - tracing was not enabled")


def _show_status(target_dir: Path, settings_file: Path) -> None:
    """Show current tracing status."""
    click.echo(f"📍 Claude tracing status in: {target_dir}")

    status = get_tracing_status(settings_file)

    if not status.enabled:
        click.echo("❌ Claude tracing is not enabled")
        if status.reason:
            click.echo(f"   Reason: {status.reason}")
        return

    click.echo("✅ Claude tracing is ENABLED")
    click.echo(f"📊 Tracking URI: {status.tracking_uri}")

    if status.experiment_id:
        click.echo(f"🔬 Experiment ID: {status.experiment_id}")
    elif status.experiment_name:
        click.echo(f"🔬 Experiment Name: {status.experiment_name}")
    else:
        click.echo("🔬 Experiment: Default (experiment 0)")


def _show_setup_status(
    target_dir: Path,
    tracking_uri: str | None,
    experiment_id: str | None,
    experiment_name: str | None,
) -> None:
    """Show setup completion status."""
    current_dir = Path.cwd().resolve()

    click.echo("\n" + "=" * 50)
    click.echo("🎯 Claude Tracing Setup Complete!")
    click.echo("=" * 50)

    click.echo(f"📁 Directory: {target_dir}")

    # Show tracking configuration
    if tracking_uri:
        click.echo(f"📊 Tracking URI: {tracking_uri}")

    if experiment_id:
        click.echo(f"🔬 Experiment ID: {experiment_id}")
    elif experiment_name:
        click.echo(f"🔬 Experiment Name: {experiment_name}")
    else:
        click.echo("🔬 Experiment: Default (experiment 0)")

    # Show next steps
    click.echo("\n" + "=" * 30)
    click.echo("🚀 Next Steps:")
    click.echo("=" * 30)

    # Only show cd if it's a different directory
    if target_dir != current_dir:
        click.echo(f"cd {target_dir}")

    click.echo("claude -p 'your prompt here'")

    if tracking_uri and tracking_uri.startswith("file://"):
        click.echo("\n💡 View your traces:")
        click.echo(f"   mlflow ui --backend-store-uri {tracking_uri}")
    elif not tracking_uri:
        click.echo("\n💡 View your traces:")
        click.echo("   mlflow ui")
    elif tracking_uri == "databricks":
        click.echo("\n💡 View your traces in your Databricks workspace")

    click.echo("\n🔧 To disable tracing later:")
    click.echo("   mlflow autolog claude --disable")



from mlflow.utils._spark_utils import _SparkDirectoryDistributor


class SparkModelCache:
    """Caches models in memory on Spark Executors, to avoid continually reloading from disk.

    This class has to be part of a different module than the one that _uses_ it. This is
    because Spark will pickle classes that are defined in the local scope, but relies on
    Python's module loading behavior for classes in different modules. In this case, we
    are relying on the fact that Python will load a module at-most-once, and can therefore
    store per-process state in a static map.
    """

    # Map from unique name --> (loaded model, local_model_path).
    _models = {}

    # Number of cache hits we've had, for testing purposes.
    _cache_hits = 0

    def __init__(self):
        pass

    @staticmethod
    def add_local_model(spark, model_path):
        """Given a SparkSession and a model_path which refers to a pyfunc directory locally,
        we will zip the directory up, enable it to be distributed to executors, and return
        the "archive_path", which should be used as the path in get_or_load().
        """
        return _SparkDirectoryDistributor.add_dir(spark, model_path)

    @staticmethod
    def get_or_load(archive_path):
        """Given a path returned by add_local_model(), this method will return a tuple of
        (loaded_model, local_model_path).
        If this Python process ever loaded the model before, we will reuse that copy.
        """
        if archive_path in SparkModelCache._models:
            SparkModelCache._cache_hits += 1
            return SparkModelCache._models[archive_path]

        local_model_dir = _SparkDirectoryDistributor.get_or_extract(archive_path)

        # We must rely on a supposed cyclic import here because we want this behavior
        # on the Spark Executors (i.e., don't try to pickle the load_model function).
        from mlflow.pyfunc import load_model

        SparkModelCache._models[archive_path] = (load_model(local_model_dir), local_model_dir)
        return SparkModelCache._models[archive_path]



# A special tag in RegisteredModel to indicate that it is a prompt
import re

IS_PROMPT_TAG_KEY = "mlflow.prompt.is_prompt"
# A special tag in ModelVersion to store the prompt text
PROMPT_TEXT_TAG_KEY = "mlflow.prompt.text"

LINKED_PROMPTS_TAG_KEY = "mlflow.linkedPrompts"

# Unity Catalog tags cannot contain dots
PROMPT_TYPE_TAG_KEY = "_mlflow_prompt_type"
RESPONSE_FORMAT_TAG_KEY = "_mlflow_prompt_response_format"

# Prompt types
PROMPT_TYPE_TEXT = "text"
PROMPT_TYPE_CHAT = "chat"

# A special tag to store associated run IDs for prompts
PROMPT_ASSOCIATED_RUN_IDS_TAG_KEY = "mlflow.prompt.associatedRunIds"

PROMPT_TEMPLATE_VARIABLE_PATTERN = re.compile(
    r"\{\{\s*([a-zA-Z_][a-zA-Z0-9_]*(?:\.[a-zA-Z_][a-zA-Z0-9_]*)*)\s*\}\}"
)

PROMPT_TEXT_DISPLAY_LIMIT = 30

# Alphanumeric, underscore, hyphen, and dot are allowed in prompt name
PROMPT_NAME_RULE = re.compile(r"^[a-zA-Z0-9_.-]+$")
